from __future__ import annotations

import copy
import dataclasses
import logging
import os
import platform
from pathlib import Path
from typing import Dict, Optional, Tuple, Any, Set, TYPE_CHECKING

import dpath.util

from checkov.common.checks_infra.registry import get_graph_checks_registry
from checkov.common.graph.checks_infra.registry import BaseRegistry
from checkov.common.typing import LibraryGraphConnector
from checkov.common.graph.graph_builder.consts import GraphSource
from checkov.common.images.image_referencer import ImageReferencerMixin
from checkov.common.output.extra_resource import ExtraResource
from checkov.common.parallelizer.parallel_runner import parallel_runner
from checkov.common.models.enums import CheckResult
from checkov.common.output.graph_record import GraphRecord
from checkov.common.output.record import Record
from checkov.common.output.report import Report, merge_reports, remove_duplicate_results
from checkov.common.bridgecrew.check_type import CheckType
from checkov.common.runners.base_runner import BaseRunner, CHECKOV_CREATE_GRAPH
from checkov.common.util import data_structures_utils
from checkov.common.util.consts import RESOLVED_MODULE_ENTRY_NAME
from checkov.common.util.parser_utils import get_module_from_full_path, get_abs_path, \
    get_tf_definition_key_from_module_dependency, TERRAFORM_NESTED_MODULE_PATH_PREFIX, \
    TERRAFORM_NESTED_MODULE_PATH_ENDING, TERRAFORM_NESTED_MODULE_PATH_SEPARATOR_LENGTH
from checkov.common.util.secrets import omit_secret_value_from_checks, omit_secret_value_from_graph_checks
from checkov.common.variables.context import EvaluationContext
from checkov.runner_filter import RunnerFilter
from checkov.terraform.checks.data.registry import data_registry
from checkov.terraform.checks.module.registry import module_registry
from checkov.terraform.checks.provider.registry import provider_registry
from checkov.terraform.checks.resource.registry import resource_registry
from checkov.terraform.checks.utils.dependency_path_handler import PATH_SEPARATOR
from checkov.terraform.context_parsers.registry import parser_registry
from checkov.terraform.evaluation.base_variable_evaluation import BaseVariableEvaluation
from checkov.common.graph.graph_builder.graph_components.attribute_names import CustomAttributes
from checkov.terraform.graph_builder.graph_components.block_types import BlockType
from checkov.terraform.graph_builder.graph_to_tf_definitions import convert_graph_vertices_to_tf_definitions
from checkov.terraform.graph_builder.local_graph import TerraformLocalGraph
from checkov.terraform.graph_manager import TerraformGraphManager
from checkov.terraform.image_referencer.manager import TerraformImageReferencerManager
from checkov.terraform.parser import Parser
from checkov.terraform.plan_utils import get_resource_id_without_nested_modules
from checkov.terraform.tag_providers import get_resource_tags
from checkov.common.runners.base_runner import strtobool

if TYPE_CHECKING:
    from networkx import DiGraph
    from checkov.common.images.image_referencer import Image

# Allow the evaluation of empty variables
dpath.options.ALLOW_EMPTY_STRING_KEYS = True

CHECK_BLOCK_TYPES = frozenset(['resource', 'data', 'provider', 'module'])


class Runner(ImageReferencerMixin[None], BaseRunner[TerraformGraphManager]):
    check_type = CheckType.TERRAFORM  # noqa: CCE003  # a static attribute

    def __init__(
        self,
        parser: Parser | None = None,
        db_connector: LibraryGraphConnector | None = None,
        external_registries: list[BaseRegistry] | None = None,
        source: str = GraphSource.TERRAFORM,
        graph_class: type[TerraformLocalGraph] = TerraformLocalGraph,
        graph_manager: TerraformGraphManager | None = None
    ) -> None:
        super().__init__(file_extensions=['.tf', '.hcl'])
        self.external_registries = [] if external_registries is None else external_registries
        self.graph_class = graph_class
        self.parser = parser or Parser()
        self.definitions: "dict[str, dict[str, Any]] | None" = None
        self.context = None
        self.breadcrumbs = None
        self.evaluations_context: Dict[str, Dict[str, EvaluationContext]] = {}
        self.graph_manager: TerraformGraphManager = graph_manager if graph_manager is not None else TerraformGraphManager(
            source=source,
            db_connector=db_connector or self.db_connector,
        )
        self.graph_registry = get_graph_checks_registry(self.check_type)
        self.definitions_with_modules: dict[str, dict[str, Any]] = {}
        self.referrer_cache: Dict[str, str] = {}
        self.non_referred_cache: Set[str] = set()
        self.enable_nested_modules = strtobool(os.getenv('CHECKOV_ENABLE_NESTED_MODULES', 'False'))

    block_type_registries = {  # noqa: CCE003  # a static attribute
        'resource': resource_registry,
        'data': data_registry,
        'provider': provider_registry,
        'module': module_registry,
    }

    def run(
            self,
            root_folder: str,
            external_checks_dir: list[str] | None = None,
            files: list[str] | None = None,
            runner_filter: RunnerFilter | None = None,
            collect_skip_comments: bool = True
    ) -> Report | list[Report]:
        runner_filter = runner_filter or RunnerFilter()
        if not runner_filter.show_progress_bar:
            self.pbar.turn_off_progress_bar()

        report = Report(self.check_type)
        parsing_errors: dict[str, Exception] = {}
        self.load_external_checks(external_checks_dir)
        local_graph = None

        if self.context is None or self.definitions is None or self.breadcrumbs is None:
            self.definitions = {}
            logging.info("Scanning root folder and producing fresh tf_definitions and context")
            if root_folder:
                root_folder = os.path.abspath(root_folder)

                local_graph, self.definitions = self.graph_manager.build_graph_from_source_directory(
                    source_dir=root_folder,
                    local_graph_class=self.graph_class,
                    download_external_modules=runner_filter.download_external_modules,
                    external_modules_download_path=runner_filter.external_modules_download_path,
                    parsing_errors=parsing_errors,
                    excluded_paths=runner_filter.excluded_paths,
                    vars_files=runner_filter.var_files,
                    create_graph=CHECKOV_CREATE_GRAPH,
                )
            elif files:
                files = [os.path.abspath(file) for file in files]
                root_folder = os.path.split(os.path.commonprefix(files))[0]
                self.parser.evaluate_variables = False
                self._parse_files(files, parsing_errors)

                if CHECKOV_CREATE_GRAPH:
                    local_graph = self.graph_manager.build_graph_from_definitions(self.definitions)
            else:
                raise Exception("Root directory was not specified, files were not specified")

            if CHECKOV_CREATE_GRAPH and local_graph:
                for vertex in local_graph.vertices:
                    if vertex.block_type == BlockType.RESOURCE:
                        if self.enable_nested_modules:
                            vertex_id = vertex.attributes.get(CustomAttributes.TF_RESOURCE_ADDRESS)
                        else:
                            vertex_id = vertex.id
                        report.add_resource(f'{vertex.path}:{vertex_id}')
                self.graph_manager.save_graph(local_graph)
                self.definitions, self.breadcrumbs = convert_graph_vertices_to_tf_definitions(
                    local_graph.vertices,
                    root_folder,
                )
        else:
            logging.info("Scanning root folder using existing tf_definitions")

        self.pbar.initiate(len(self.definitions))
        self.check_tf_definition(report, root_folder, runner_filter, collect_skip_comments)

        report.add_parsing_errors(parsing_errors.keys())

        if CHECKOV_CREATE_GRAPH:
            graph_report = self.get_graph_checks_report(root_folder, runner_filter)
            merge_reports(report, graph_report)

        report = remove_duplicate_results(report)

        if runner_filter.run_image_referencer:
            image_report = self.check_container_image_references(
                graph_connector=self.graph_manager.get_reader_endpoint(),
                root_path=root_folder,
                runner_filter=runner_filter,
            )

            if image_report:
                # due too many tests failing only return a list, if there is an image report
                return [report, image_report]

        return report

    def load_external_checks(self, external_checks_dir: list[str] | None) -> None:
        if external_checks_dir:
            for directory in external_checks_dir:
                resource_registry.load_external_checks(directory)
                self.graph_registry.load_external_checks(directory)

    def get_connected_node(self, entity, root_folder) -> Optional[Dict[str, Any]]:
        connected_entity = entity.get('connected_node')
        if not connected_entity:
            return None
        connected_entity_context, connected_entity_evaluations = self.get_entity_context_and_evaluations(
            connected_entity
        )
        if not connected_entity_context:
            return None
        full_file_path = connected_entity[CustomAttributes.FILE_PATH]
        connected_node_data = {}
        connected_node_data['code_block'] = connected_entity_context.get('code_lines')
        connected_node_data['file_path'] = f"/{os.path.relpath(full_file_path, root_folder)}"
        connected_node_data['file_line_range'] = [connected_entity_context.get('start_line'),
                                                  connected_entity_context.get('end_line')]
        connected_node_data['resource'] = ".".join(connected_entity_context['definition_path'])
        connected_node_data['entity_tags'] = connected_entity.get('tags', {})
        connected_node_data['evaluations'] = connected_entity_evaluations
        connected_node_data['file_abs_path'] = os.path.abspath(full_file_path)
        connected_node_data['resource_address'] = connected_entity_context.get('address')
        return connected_node_data

    def get_graph_checks_report(self, root_folder: str, runner_filter: RunnerFilter) -> Report:
        report = Report(self.check_type)
        checks_results = self.run_graph_checks_results(runner_filter, self.check_type)

        for check, check_results in checks_results.items():
            for check_result in check_results:
                entity = check_result['entity']
                entity_context, entity_evaluations = self.get_entity_context_and_evaluations(entity)
                if entity_context:
                    full_file_path = entity[CustomAttributes.FILE_PATH]
                    copy_of_check_result = copy.deepcopy(check_result)
                    for skipped_check in entity_context.get('skipped_checks', []):
                        if skipped_check['id'] == check.id:
                            copy_of_check_result['result'] = CheckResult.SKIPPED
                            copy_of_check_result['suppress_comment'] = skipped_check['suppress_comment']
                            break
                    copy_of_check_result['entity'] = entity.get(CustomAttributes.CONFIG)
                    connected_node_data = self.get_connected_node(entity, root_folder)
                    if platform.system() == "Windows":
                        root_folder = os.path.split(full_file_path)[0]
                    resource_id = ".".join(entity_context['definition_path'])
                    resource = resource_id
                    module_dependency = entity.get(CustomAttributes.MODULE_DEPENDENCY)
                    module_dependency_num = entity.get(CustomAttributes.MODULE_DEPENDENCY_NUM)
                    definition_context_file_path = full_file_path
                    if module_dependency and module_dependency_num:
                        if self.enable_nested_modules:
                            resource = entity.get(CustomAttributes.TF_RESOURCE_ADDRESS, resource_id)
                        else:
                            module_dependency_path = module_dependency.split(PATH_SEPARATOR)[-1]
                            tf_path = get_tf_definition_key_from_module_dependency(full_file_path, module_dependency_path, module_dependency_num)
                            referrer_id = self._find_id_for_referrer(tf_path)
                            if referrer_id:
                                resource = f'{referrer_id}.{resource_id}'
                        definition_context_file_path = get_tf_definition_key_from_module_dependency(full_file_path, module_dependency, module_dependency_num)
                    elif entity.get(CustomAttributes.TF_RESOURCE_ADDRESS) and entity.get(CustomAttributes.TF_RESOURCE_ADDRESS) != resource_id:
                        # for plan resources
                        resource = entity[CustomAttributes.TF_RESOURCE_ADDRESS]
                        if not self.enable_nested_modules:
                            resource = get_resource_id_without_nested_modules(resource)
                    entity_config = self.get_graph_resource_entity_config(entity)
                    censored_code_lines = omit_secret_value_from_graph_checks(
                        check=check,
                        check_result=check_result,
                        entity_code_lines=entity_context.get('code_lines'),
                        entity_config=entity_config,
                        resource_attributes_to_omit=runner_filter.resource_attr_to_omit
                    )
                    record = Record(
                        check_id=check.id,
                        bc_check_id=check.bc_id,
                        check_name=check.name,
                        check_result=copy_of_check_result,
                        code_block=censored_code_lines,
                        file_path=f"/{os.path.relpath(full_file_path, root_folder)}",
                        file_line_range=[entity_context.get('start_line'),
                                         entity_context.get('end_line')],
                        resource=resource,
                        entity_tags=entity.get('tags', {}),
                        evaluations=entity_evaluations,
                        check_class=check.__class__.__module__,
                        file_abs_path=os.path.abspath(full_file_path),
                        resource_address=entity_context.get('address'),
                        severity=check.severity,
                        bc_category=check.bc_category,
                        benchmarks=check.benchmarks,
                        connected_node=connected_node_data,
                        definition_context_file_path=definition_context_file_path
                    )
                    if self.breadcrumbs:
                        if self.enable_nested_modules:
                            breadcrumb = self.breadcrumbs.get(record.file_path, {}).get(resource)
                        else:
                            breadcrumb = self.breadcrumbs.get(record.file_path, {}).get(resource_id)
                        if breadcrumb:
                            record = GraphRecord(record, breadcrumb)
                    record.set_guideline(check.guideline)
                    report.add_record(record=record)
        return report

    def get_entity_context_and_evaluations(self, entity):
        entity_evaluations = None
        block_type = entity[CustomAttributes.BLOCK_TYPE]
        full_file_path = entity[CustomAttributes.FILE_PATH]
        if entity.get(CustomAttributes.MODULE_DEPENDENCY):
            full_file_path = get_tf_definition_key_from_module_dependency(full_file_path, entity[CustomAttributes.MODULE_DEPENDENCY], entity[CustomAttributes.MODULE_DEPENDENCY_NUM])
        definition_path = entity[CustomAttributes.BLOCK_NAME].split('.')
        entity_context_path = [block_type] + definition_path
        entity_context = self.context.get(full_file_path, {})
        try:
            for k in entity_context_path:
                if k in entity_context:
                    entity_context = entity_context[k]
                else:
                    logging.warning(f'Failed to find context for {".".join(entity_context_path)}')
                    return None, None
            entity_context['definition_path'] = definition_path
        except StopIteration:
            logging.debug(f"Did not find context for key {full_file_path}")
        return entity_context, entity_evaluations

    def check_tf_definition(self, report: Report, root_folder: Path, runner_filter: RunnerFilter,
                            collect_skip_comments=True) -> None:
        parser_registry.reset_definitions_context()
        if not self.context:
            definitions_context = {}
            for definition in self.definitions.items():
                definitions_context = parser_registry.enrich_definitions_context(definition, collect_skip_comments)
            self.context = definitions_context
            logging.debug('Created definitions context')

        if self.enable_nested_modules:
            self.push_skipped_checks_down_from_modules(self.context)
        for full_file_path, definition in self.definitions.items():
            self.pbar.set_additional_data({'Current File Scanned': os.path.relpath(full_file_path, root_folder)})
            if self.enable_nested_modules:
                abs_scanned_file = get_abs_path(full_file_path)
                abs_referrer = None
            else:
                abs_scanned_file, abs_referrer = self._strip_module_referrer(full_file_path)
            scanned_file = f"/{os.path.relpath(abs_scanned_file, root_folder)}"
            logging.debug(f"Scanning file: {scanned_file}")
            self.run_all_blocks(definition, self.context, full_file_path, root_folder, report,
                                scanned_file, runner_filter, abs_referrer)
            self.pbar.update()
        self.pbar.close()

    def run_all_blocks(self, definition, definitions_context, full_file_path, root_folder, report,
                       scanned_file, runner_filter, module_referrer: Optional[str]):
        if not definition:
            logging.debug("Empty definition, skipping run (root_folder=%s)", root_folder)
            return
        block_types = set(definition.keys())
        for block_type in block_types & CHECK_BLOCK_TYPES:
            self.run_block(definition[block_type], definitions_context,
                           full_file_path, root_folder, report,
                           scanned_file, block_type, runner_filter, None, module_referrer)

    def run_block(self, entities,
                  definition_context,
                  full_file_path, root_folder, report, scanned_file,
                  block_type, runner_filter=None, entity_context_path_header=None,
                  module_referrer: Optional[str] = None):

        registry = self.block_type_registries[block_type]
        if not registry:
            return

        for entity in entities:
            entity_evaluations = None
            context_parser = parser_registry.context_parsers[block_type]
            definition_path = context_parser.get_entity_context_path(entity)
            (entity_type, entity_name, entity_config) = registry.extract_entity_details(entity)
            entity_id = ".".join(definition_path)  # example: aws_s3_bucket.my_bucket

            caller_file_path = None
            caller_file_line_range = None

            if self.enable_nested_modules:
                entity_id = entity_config.get(CustomAttributes.TF_RESOURCE_ADDRESS)
                module, _ = get_module_from_full_path(full_file_path)
                if module:
                    full_definition_path = entity_id.split('.')
                    module_name_index = len(full_definition_path) - full_definition_path[::-1].index(BlockType.MODULE)  # the next item after the last 'module' prefix is the module name
                    module_name = full_definition_path[module_name_index]
                    caller_context = definition_context[module].get(BlockType.MODULE, {}).get(module_name)
                    caller_file_line_range = [caller_context.get('start_line'), caller_context.get('end_line')]
                    abs_caller_file = get_abs_path(module)
                    caller_file_path = f"/{os.path.relpath(abs_caller_file, root_folder)}"
            elif module_referrer is not None:
                referrer_id = self._find_id_for_referrer(full_file_path)

                if referrer_id:
                    entity_id = f"{referrer_id}.{entity_id}"  # ex: module.my_module.aws_s3_bucket.my_bucket
                    abs_caller_file = module_referrer[:module_referrer.rindex("#")]
                    caller_file_path = f"/{os.path.relpath(abs_caller_file, root_folder)}"

                    try:
                        caller_context = definition_context[abs_caller_file]
                        for part in referrer_id.split("."):
                            caller_context = caller_context[part]
                    except KeyError:
                        logging.debug("Unable to find caller context for: %s", abs_caller_file)
                        caller_context = None

                    if caller_context:
                        caller_file_line_range = [caller_context.get('start_line'), caller_context.get('end_line')]
                else:
                    logging.debug(f"Unable to find referrer ID for full path: {full_file_path}")

            if entity_context_path_header is None:
                entity_context_path = [block_type] + definition_path
            else:
                entity_context_path = entity_context_path_header + block_type + definition_path
            # Entity can exist only once per dir, for file as well
            try:
                entity_context = data_structures_utils.get_inner_dict(
                    definition_context[full_file_path],
                    entity_context_path,
                )
                entity_lines_range = [entity_context.get('start_line'), entity_context.get('end_line')]
                entity_code_lines = entity_context.get('code_lines')
                skipped_checks = entity_context.get('skipped_checks')
            except KeyError:
                # TODO: Context info isn't working for modules
                entity_lines_range = None
                entity_code_lines = None
                skipped_checks = None

            if not self.enable_nested_modules and block_type == "module":
                self.push_skipped_checks_down_old(definition_context, full_file_path, skipped_checks)

            if full_file_path in self.evaluations_context:
                variables_evaluations = {}
                for var_name, context_info in self.evaluations_context.get(full_file_path, {}).items():
                    variables_evaluations[var_name] = dataclasses.asdict(context_info)
                entity_evaluations = BaseVariableEvaluation.reduce_entity_evaluations(variables_evaluations,
                                                                                      entity_context_path)
            results = registry.scan(scanned_file, entity, skipped_checks, runner_filter)
            absolut_scanned_file_path, _ = self._strip_module_referrer(file_path=full_file_path)
            # This duplicates a call at the start of scan, but adding this here seems better than kludging with some tuple return type
            tags = get_resource_tags(entity_type, entity_config)
            if results:
                for check, check_result in results.items():
                    censored_code_lines = omit_secret_value_from_checks(
                        check=check,
                        check_result=check_result,
                        entity_code_lines=entity_code_lines,
                        entity_config=entity_config,
                        resource_attributes_to_omit=runner_filter.resource_attr_to_omit
                    )
                    record = Record(
                        check_id=check.id,
                        bc_check_id=check.bc_id,
                        check_name=check.name,
                        check_result=check_result,
                        code_block=censored_code_lines,
                        file_path=scanned_file,
                        file_line_range=entity_lines_range,
                        resource=entity_id,
                        evaluations=entity_evaluations,
                        check_class=check.__class__.__module__,
                        file_abs_path=absolut_scanned_file_path,
                        entity_tags=tags,
                        caller_file_path=caller_file_path,
                        caller_file_line_range=caller_file_line_range,
                        severity=check.severity,
                        bc_category=check.bc_category,
                        benchmarks=check.benchmarks,
                        details=check.details,
                        definition_context_file_path=full_file_path
                    )
                    if CHECKOV_CREATE_GRAPH:
                        if self.enable_nested_modules:
                            entity_key = entity_id
                        else:
                            entity_key = f"{entity_type}.{entity_name}"
                        breadcrumb = self.breadcrumbs.get(record.file_path, {}).get(entity_key)
                        if breadcrumb:
                            record = GraphRecord(record, breadcrumb)

                    record.set_guideline(check.guideline)
                    report.add_record(record=record)
            else:
                if block_type == "resource":
                    # resources without checks, but not existing ones
                    report.extra_resources.add(
                        ExtraResource(
                            file_abs_path=absolut_scanned_file_path,
                            file_path=scanned_file,
                            resource=entity_id,
                        )
                    )

    def _parse_files(self, files, parsing_errors):
        def parse_file(file):
            if not (file.endswith(".tf") or file.endswith(".hcl")):
                return
            file_parsing_errors = {}
            parse_result = self.parser.parse_file(file=file, parsing_errors=file_parsing_errors)
            # the exceptions type can un-pickleable so we need to cast them to Exception
            for path, e in file_parsing_errors.items():
                file_parsing_errors[path] = Exception(e.__repr__())
            return file, parse_result, file_parsing_errors

        results = parallel_runner.run_function(parse_file, files)
        for result in results:
            if result:
                file, parse_result, file_parsing_errors = result
                if parse_result is not None:
                    self.definitions[file] = parse_result
                if file_parsing_errors:
                    parsing_errors.update(file_parsing_errors)

    @staticmethod
    def push_skipped_checks_down_old(definition_context, module_path, skipped_checks):
        # this method pushes the skipped_checks down the 1 level to all resource types.

        if skipped_checks is None:
            return

        if len(skipped_checks) == 0:
            return

        # iterate over definitions to find those that reference the module path
        # definition is in the format <file>[<referrer>#<index>]
        # where referrer could be a path, or path1->path2, etc

        for definition in definition_context:
            _, mod_ref = Runner._strip_module_referrer(definition)
            if mod_ref is None:
                continue

            if module_path not in mod_ref:
                continue

            for block_type, block_configs in definition_context[definition].items():
                # skip if type is not a Terraform resource
                if block_type not in CHECK_BLOCK_TYPES:
                    continue

                if block_type == "module":
                    # modules don't have a type, just a name
                    for resource_config in block_configs.values():
                        # append the skipped checks also from a module to another module
                        resource_config["skipped_checks"] += skipped_checks
                else:
                    # there may be multiple resource types - aws_bucket, etc
                    for resource_configs in block_configs.values():
                        # there may be multiple names for each resource type
                        for resource_config in resource_configs.values():
                            # append the skipped checks from the module to the other resources.
                            resource_config["skipped_checks"] += skipped_checks

    def push_skipped_checks_down_from_modules(self, definition_context):
        module_context_parser = parser_registry.context_parsers[BlockType.MODULE]
        for full_file_path, definition in self.definitions.items():
            definition_modules_context = definition_context.get(full_file_path, {}).get(BlockType.MODULE, {})
            for entity in definition.get(BlockType.MODULE, []):
                module_name = module_context_parser.get_entity_context_path(entity)[0]
                skipped_checks = definition_modules_context.get(module_name).get('skipped_checks')
                resolved_paths = entity.get(module_name).get(RESOLVED_MODULE_ENTRY_NAME)
                self.push_skipped_checks_down(definition_context, skipped_checks, resolved_paths)

    def push_skipped_checks_down(self, definition_context, skipped_checks, resolved_paths):
        # this method pushes the skipped_checks down the 1 level to all resource types.
        if not skipped_checks:
            return

        for definition in resolved_paths:
            for block_type, block_configs in definition_context.get(definition, {}).items():
                # skip if type is not a Terraform resource
                if block_type not in CHECK_BLOCK_TYPES:
                    continue

                if block_type == "module":
                    # modules don't have a type, just a name
                    for module_name, module_config in block_configs.items():
                        # append the skipped checks also from a module to another module
                        module_config["skipped_checks"] += skipped_checks
                        module_context = next(m for m in self.definitions.get(definition).get(block_type) if module_name in m)
                        recursive_resolved_paths = module_context.get(module_name).get(RESOLVED_MODULE_ENTRY_NAME)
                        self.push_skipped_checks_down(definition_context, skipped_checks, recursive_resolved_paths)
                else:
                    # there may be multiple resource types - aws_bucket, etc
                    for resource_configs in block_configs.values():
                        # there may be multiple names for each resource type
                        for resource_config in resource_configs.values():
                            # append the skipped checks from the module to the other resources.
                            resource_config["skipped_checks"] += skipped_checks

    @staticmethod
    def _strip_module_referrer(file_path: str) -> Tuple[str, Optional[str]]:
        """
        For file paths containing module referrer information (e.g.: "module/module.tf[main.tf#0]"), this
        returns a tuple containing the file path (e.g., "module/module.tf") and referrer (e.g., "main.tf#0").
        If the file path does not contain a referred, the tuple will contain the original file path and None.
        """
        if file_path.endswith(TERRAFORM_NESTED_MODULE_PATH_ENDING) and TERRAFORM_NESTED_MODULE_PATH_PREFIX in file_path:
            return file_path[:file_path.index(TERRAFORM_NESTED_MODULE_PATH_PREFIX)], \
                file_path[file_path.index(TERRAFORM_NESTED_MODULE_PATH_PREFIX) + TERRAFORM_NESTED_MODULE_PATH_SEPARATOR_LENGTH: -TERRAFORM_NESTED_MODULE_PATH_SEPARATOR_LENGTH]
        else:
            return file_path, None

    def _find_id_for_referrer(self, full_file_path) -> Optional[str]:
        cached_referrer = self.referrer_cache.get(full_file_path)
        if cached_referrer:
            return cached_referrer
        if full_file_path in self.non_referred_cache:
            return None

        if not self.definitions_with_modules:
            self._prepare_definitions_with_modules()
        for file_content in self.definitions_with_modules.values():
            for modules in file_content["module"]:
                for module_name, module_content in modules.items():
                    if RESOLVED_MODULE_ENTRY_NAME not in module_content:
                        continue

                    if full_file_path in module_content[RESOLVED_MODULE_ENTRY_NAME]:
                        id_referrer = f"module.{module_name}"
                        self.referrer_cache[full_file_path] = id_referrer
                        return id_referrer

        self.non_referred_cache.add(full_file_path)
        return None

    def _prepare_definitions_with_modules(self) -> None:
        def __cache_file_content(file_name: str, file_modules: list[dict[str, Any]]) -> None:
            for modules in file_modules:
                for module_content in modules.values():
                    if RESOLVED_MODULE_ENTRY_NAME in module_content:
                        self.definitions_with_modules[file_name] = file_content
                        return

        for file, file_content in self.definitions.items():
            if "module" in file_content:
                __cache_file_content(file_name=file, file_modules=file_content["module"])

    def extract_images(
        self,
        graph_connector: DiGraph | None = None,
        definitions: dict[str, dict[str, Any] | list[dict[str, Any]]] | None = None,
        definitions_raw: dict[str, list[tuple[int, str]]] | None = None
    ) -> list[Image]:
        if not graph_connector:
            # should not happen
            return []

        manager = TerraformImageReferencerManager(graph_connector=graph_connector)
        images = manager.extract_images_from_resources()

        return images

    @staticmethod
    def get_graph_resource_entity_config(entity):
        context_parser = parser_registry.context_parsers[entity[CustomAttributes.BLOCK_TYPE]]
        entity_config = entity[CustomAttributes.CONFIG]
        definition_path = context_parser.get_entity_definition_path(entity_config)
        for path in definition_path:
            entity_config = entity_config[path]
        return entity_config



import itertools
import logging
import re
from datetime import datetime, timedelta
from functools import reduce
from math import ceil, floor, log
from typing import Union, Any, Dict, Callable, List, Optional

from checkov.terraform.parser_functions import tonumber, FUNCTION_FAILED, create_map, tobool, tostring

"""
This file contains a custom implementation of the builtin `eval` function.
`eval` is not a safe function, because it can execute *every* command,
so this file overrides `eval` and allows only the functions in SAFE_EVAL_DICT.

The functions are an implementation of Terraform's built-in functions
https://www.terraform.io/docs/configuration/functions.html

Not all of the functions are implemented yet. If a function doesn't exist, the original value is returned.
"""


def _find_regex_groups(pattern: str, input_str: str) -> Optional[Union[Dict[str, str], List[str]]]:
    match = re.match(pattern, input_str)
    if match:
        if match.groupdict():
            # try to find named capturing groups
            return match.groupdict()
        if list(match.groups()):
            # try to find unnamed capturing groups
            return list(match.groups())
    return None


def regex(pattern: str, input_str: str) -> Union[Dict[str, str], List[str], str]:
    try:
        groups = _find_regex_groups(pattern, input_str)
        if groups is not None:
            return groups

        results: List[str] = re.findall(pattern, input_str)
        # return first match
        if len(results) > 0:
            return results[0]
        return ""
    except TypeError:
        return f"regex({pattern}, {input_str})"


def regexall(pattern: str, input_str: str) -> Union[Dict[str, str], List[str], str]:
    try:
        groups = _find_regex_groups(pattern, input_str)
        if groups is not None:
            return groups

        results = re.findall(pattern, input_str)
        return results
    except TypeError:
        return f"regexall({pattern}, {input_str})"


def trim(input_str: str, chars_to_remove: str) -> str:
    for c in chars_to_remove:
        input_str = input_str.replace(c, "")
    return input_str


def coalesce(*arg: Any) -> Any:
    return reduce(lambda x, y: x if x not in [None, ""] else y, arg)


def coalesce_list(*arg: List[Any]) -> List[Any]:
    return reduce(lambda x, y: x if x not in [None, []] else y, arg)


def flatten(lst: List[List[Any]]) -> List[Any]:
    res = [item for sublist in lst for item in sublist]
    if any(type(elem) is list for elem in res):
        return flatten(res)
    else:
        return res


def matchkeys(values_list: List[Any], keys_list: List[Any], search_set: List[Any]) -> List[Any]:
    matching = set()
    for search in search_set:
        indices = [i for i, x in enumerate(keys_list) if x == search]
        for i in indices:
            matching.add(values_list[i])

    return list(matching)


def reverse(lst: List[Any]) -> List[Any]:
    lst.reverse()
    return lst


def sort(lst: List[str]) -> List[str]:
    lst.sort()
    return lst


def merge(*args: Any) -> Dict[str, Any]:
    res: Dict[str, Any] = {}
    for d in args:
        res = {**res, **d}
    return res


def wrap_func(f: Callable[..., Any], *args: Any) -> Any:
    res = f(*args)
    if res == FUNCTION_FAILED:
        raise ValueError
    return res


def update_datetime(dt: datetime, delta: timedelta, adding: bool) -> datetime:
    if adding is True:
        dt = dt + delta
    else:
        dt = dt - delta
    return dt


def timeadd(input_str: str, time_delta: str) -> str:
    '''
    From docs:
    duration is a string representation of a time difference, consisting of sequences of number and unit pairs,
     like "1.5h" or "1h30m". The accepted units are "ns", "us" (or "µs"), "ms", "s", "m", and "h".
     The first number may be negative to indicate a negative duration, like "-2h5m".
    '''

    # Convert the date to allowing parsing
    input_str = input_str.replace("Z", "+00:00")
    dt = datetime.fromisoformat(input_str)
    adding = True
    if time_delta[0] == '-':
        adding = False
        time_delta = time_delta[1:]
    # Split out into each of the deltas
    deltas = re.split(r'(\d*\.*\d+)', time_delta)
    # Needed to strip the leading empty element
    deltas = list(filter(None, deltas))
    while len(deltas) > 0:
        amount = float(deltas[0])
        interval = deltas[1]
        deltas = deltas[2:]
        delta = timedelta(0)
        if interval == 'h':
            delta = timedelta(hours=amount)
        elif interval == 'm':
            delta = timedelta(minutes=amount)
        elif interval == 's':
            delta = timedelta(seconds=amount)
        elif interval == 'ms':
            delta = timedelta(milliseconds=amount)
        elif interval == 'us' or interval == 'µs':
            delta = timedelta(microseconds=amount)
        elif interval == 'ns':  # Crude, but timedelta does not deal with nanoseconds
            delta = timedelta(microseconds=(amount / 1000))

        dt = update_datetime(dt, delta, adding)
        
    return dt.strftime('%Y-%m-%dT%H:%M:%SZ')


def process_formatting_codes(format_str: str, dt: datetime) -> str:
    format_mapping = {
        "YYYY": "%Y",
        "YY": "%y",
        "MMMM": "%B",
        "MMM": "%b",
        "MM": "%m",
        "M": "%-m",
        "DD" : "%d",
        "D" : "%-d",
        "EEEE" : "%A",
        "EEE" : "%a",
        "HH" : "%I",
        "H" : "%-I",
        "hh" : "%H",
        "h" : "%-H",
        "mm" : "%M",
        "m" : "%-M",
        "ss" : "%S",
        "s" : "%-S",
        "AA" : "%p",
        # "aa" : "%p",  # included for completeness but requires separate handling
        "ZZZZZ" : "%z",
        "ZZZZ" : "%z",
        "ZZZ" : "%z",
        "Z " : "%z"}

    if format_str == 'aa':
        format_str = dt.strftime('%p').lower()
    elif format_str == 'ZZZZZ':
        tz = dt.strftime("%z")
        format_str = tz[:3] + ":" + tz[3:]
    elif format_str == 'ZZZ':
        tz = dt.strftime("%z")
        if tz == '+0000':
            tz = 'UTC'
        format_str = tz
    elif format_str == 'Z':
        tz = dt.strftime("%z")
        if tz == '+0000':
            tz = 'Z'
        format_str = tz
    else:
        format_str = format_mapping.get(format_str, format_str)

    return format_str


def formatdate(format_str: str, input_str: str) -> str:
    '''
    From docs: This function is intended for producing common machine-oriented timestamp formats such as
    those defined in RFC822, RFC850, and RFC1123. It is not suitable for truly human-oriented date
    formatting because it is not locale-aware.
    Any non-letter characters, such as punctuation, are reproduced verbatim in the output.
    To include literal letters in the format string, enclose them in single quotes '.
    To include a literal quote, escape it by doubling the quotes.
    Function works through the format string halting on single quotes to process any formatting
    '''

    # Convert the input str to a date
    input_str = input_str.replace("Z", "+00:00")
    dt = datetime.fromisoformat(input_str)

    processed_format_str = ""
    format_str_segment = ""
    in_quote = False  # Keep track of whether in formatting or quoted text
    last_ch = ""  # Used to identify the '' scenario
    for ch in format_str:
        if ch == "'" or in_quote is True:
            if len(format_str_segment) > 0:
                processed_format_str += process_formatting_codes(format_str_segment, dt)
                format_str_segment = ""
            if ch == "'":
                if last_ch == "'":
                    processed_format_str += "'"
                in_quote = not in_quote
            else:
                processed_format_str += ch
        else:
            if ch != last_ch and last_ch != "":  # new format code and the start of the string
                processed_format_str += process_formatting_codes(format_str_segment, dt)
                format_str_segment = ""
            format_str_segment += ch
        last_ch = ch
    if len(format_str_segment) > 0:
        processed_format_str += process_formatting_codes(format_str_segment, dt)

    return dt.strftime(processed_format_str)


SAFE_EVAL_FUNCTIONS: List[str] = []
SAFE_EVAL_DICT = dict([(k, locals().get(k, None)) for k in SAFE_EVAL_FUNCTIONS])

# math functions
SAFE_EVAL_DICT["abs"] = abs
SAFE_EVAL_DICT["ceil"] = ceil
SAFE_EVAL_DICT["floor"] = floor
SAFE_EVAL_DICT["log"] = log
SAFE_EVAL_DICT["max"] = max
SAFE_EVAL_DICT["min"] = min
SAFE_EVAL_DICT["parsint"] = int
SAFE_EVAL_DICT["pow"] = pow
SAFE_EVAL_DICT["signum"] = lambda x: -1 if x < 0 else 0 if x == 0 else 1

# string functions
SAFE_EVAL_DICT["chomp"] = lambda x: x.rstrip()
SAFE_EVAL_DICT["format"] = lambda text_to_format, *args: (text_to_format % args)
SAFE_EVAL_DICT["formatlist"] = lambda text_to_format, args_list: [(text_to_format % args) for args in args_list]
SAFE_EVAL_DICT["indent"] = lambda num_of_space, input_str: input_str
SAFE_EVAL_DICT["join"] = lambda separator, lst: separator.join(lst)
SAFE_EVAL_DICT["lower"] = lambda input_str: input_str.lower()
SAFE_EVAL_DICT["regex"] = regex
SAFE_EVAL_DICT["regexall"] = regexall
SAFE_EVAL_DICT["replace"] = lambda string, substring, replacement: string.replace(substring, replacement)
SAFE_EVAL_DICT["split"] = lambda separator, input_str: input_str.split(separator)
SAFE_EVAL_DICT["strrev"] = lambda input_str: input_str[::-1]
SAFE_EVAL_DICT["substr"] = lambda input_str, offset, length: input_str[offset : offset + length]
SAFE_EVAL_DICT["title"] = lambda input_str: input_str.title()
SAFE_EVAL_DICT["trim"] = trim
SAFE_EVAL_DICT["trimprefix"] = lambda input_str, prefix: input_str.lstrip(prefix)
SAFE_EVAL_DICT["trimsuffix"] = lambda input_str, prefix: input_str.rstrip(prefix)
SAFE_EVAL_DICT["trimspace"] = lambda input_str: input_str.strip()
SAFE_EVAL_DICT["upper"] = lambda input_str: input_str.upper()

# collections
SAFE_EVAL_DICT["chunklist"] = lambda lst, chunk_size: [lst[i : i + chunk_size] for i in range(0, len(lst), chunk_size)]
SAFE_EVAL_DICT["coalesce"] = coalesce
SAFE_EVAL_DICT["coalescelist"] = coalesce_list
SAFE_EVAL_DICT["compact"] = lambda lst: list(filter(lambda l: l != "", lst))
SAFE_EVAL_DICT["concat"] = lambda *lists: list(itertools.chain(*lists))
SAFE_EVAL_DICT["contains"] = lambda lst, value: value in lst
SAFE_EVAL_DICT["distinct"] = lambda lst: list(dict.fromkeys(lst))
SAFE_EVAL_DICT["element"] = lambda lst, index: lst[index]
SAFE_EVAL_DICT["flatten"] = flatten
SAFE_EVAL_DICT["index"] = lambda lst, value: lst.index(value)
SAFE_EVAL_DICT["keys"] = lambda map_input: list(map_input.keys())
SAFE_EVAL_DICT["length"] = len
SAFE_EVAL_DICT["list"] = lambda *args: list(args)
SAFE_EVAL_DICT["lookup"] = lambda map_input, key, default: map_input.get(key, default)
SAFE_EVAL_DICT["map"] = lambda *args: wrap_func(create_map, list(args))
SAFE_EVAL_DICT["matchkeys"] = matchkeys
SAFE_EVAL_DICT["merge"] = merge
# SAFE_EVAL_DICT['range']
SAFE_EVAL_DICT["reverse"] = reverse
SAFE_EVAL_DICT["sort"] = sort


# type conversion
SAFE_EVAL_DICT["tobool"] = lambda arg: wrap_func(tobool, arg)
SAFE_EVAL_DICT["tolist"] = lambda *args: list(*args)
# SAFE_EVAL_DICT["tomap"] = lambda arg: wrap_func(tomap, str(arg))
SAFE_EVAL_DICT["tonumber"] = lambda arg: arg if type(arg) in [int, float] else wrap_func(tonumber, arg)
SAFE_EVAL_DICT["toset"] = lambda origin: set(origin)
SAFE_EVAL_DICT["tostring"] = lambda arg: arg if isinstance(arg, str) else wrap_func(tostring, str(arg))

# encoding
SAFE_EVAL_DICT["jsonencode"] = lambda arg: arg

# date functions
SAFE_EVAL_DICT["timestamp"] = lambda: datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
SAFE_EVAL_DICT["timeadd"] = timeadd
SAFE_EVAL_DICT["formatdate"] = formatdate


def evaluate(input_str: str) -> Any:
    if "__" in input_str:
        logging.debug(f"got a substring with double underscore, which is not allowed. origin string: {input_str}")
        return input_str
    evaluated = eval(input_str, {"__builtins__": None}, SAFE_EVAL_DICT)  # nosec
    return evaluated if not isinstance(evaluated, str) else remove_unicode_null(evaluated)


def remove_unicode_null(input_str: str) -> str:
    return input_str.replace("\u0000", "\\0")



from __future__ import annotations

from typing import Any

from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.common.models.enums import CheckResult, CheckCategories


class VPCSecurityGroupAllowAll(BaseResourceCheck):
    def __init__(self) -> None:
        name = "Ensure security group does not contain allow-all rules."
        id = "CKV_YC_19"
        categories = (CheckCategories.GENERAL_SECURITY,)
        supported_resources = ("yandex_vpc_security_group",)
        super().__init__(
            name=name,
            id=id,
            categories=categories,
            supported_resources=supported_resources,
        )

    def scan_resource_conf(self, conf: dict[str, list[Any]]) -> CheckResult:
        if "ingress" in conf.keys():
            cidr_block = conf["ingress"][0]["v4_cidr_blocks"]
            for cidr in cidr_block[0]:
                if cidr == "0.0.0.0/0":
                    if "port" in conf["ingress"][0].keys():
                        if conf["ingress"][0]["port"][0] == -1:
                            return CheckResult.FAILED
                        return CheckResult.PASSED
                    if "from_port" not in conf["ingress"][0].keys() and "to_port" not in conf["ingress"][0].keys():
                        return CheckResult.FAILED
                    if conf["ingress"][0]["from_port"][0] == 0 and conf["ingress"][0]["to_port"][0] == 65535:
                        return CheckResult.FAILED
        return CheckResult.PASSED


scanner = VPCSecurityGroupAllowAll()



from __future__ import annotations

import os

from checkov.common.graph.graph_builder import CustomAttributes
from checkov.common.images.graph.image_referencer_provider import GraphImageReferencerProvider
from checkov.common.images.image_referencer import Image
from checkov.common.util.consts import START_LINE, END_LINE
from checkov.common.util.str_utils import removeprefix


class BaseCloudFormationProvider(GraphImageReferencerProvider):

    def extract_images_from_resources(self) -> list[Image]:
        images = []

        supported_resources_graph = self.extract_nodes()

        for _, resource in supported_resources_graph.nodes(data=True):
            image_names: list[str] = []
            resource_type = resource[CustomAttributes.RESOURCE_TYPE]

            extract_images_func = self.supported_resource_types.get(resource_type)
            if extract_images_func:
                image_names.extend(extract_images_func(resource))

            for name in image_names:
                images.append(
                    Image(
                        file_path=resource[CustomAttributes.FILE_PATH],
                        name=name,
                        start_line=resource[START_LINE],
                        end_line=resource[END_LINE],
                        related_resource_id=f'{removeprefix(resource.get("file_path_"), os.getenv("BC_ROOT_DIR", ""))}:{resource.get("id_")}',
                    )
                )

        return images



from __future__ import annotations

import datetime
from typing import Any

from bc_jsonpath_ng import parse

from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.github.base_github_configuration_check import BaseGithubCheck
from checkov.github.schemas.branch import schema as branch_schema
from checkov.json_doc.enums import BlockType


class GithubDisallowInactiveBranch60Days(BaseGithubCheck):
    def __init__(self) -> None:
        name = "Ensure inactive branches are reviewed and removed periodically"
        id = "CKV_GITHUB_15"
        categories = (CheckCategories.SUPPLY_CHAIN,)
        super().__init__(
            id=id,
            name=name,
            categories=categories,
            supported_entities=("*",),
            block_type=BlockType.DOCUMENT,
        )

    def scan_entity_conf(self, conf: dict[str, Any], entity_type: str) -> CheckResult:  # type:ignore[override]
        if branch_schema.validate(conf):
            evaluated_key = self.get_evaluated_keys()[0].replace("/", ".")
            jsonpath_expression = parse(f"$..{evaluated_key}")
            matches = jsonpath_expression.find(conf)
            if matches:
                last_commit = matches[0].value.get('date', '')
                if last_commit:
                    two_months_ago = datetime.datetime.today() - datetime.timedelta(days=60)
                    last_commit_date = datetime.datetime.strptime(last_commit, "%Y-%m-%dT%H:%M:%SZ")
                    if last_commit_date < two_months_ago:
                        return CheckResult.FAILED
                    return CheckResult.PASSED
        return CheckResult.UNKNOWN

    def get_evaluated_keys(self) -> list[str]:
        return ['commit/commit/author']


check = GithubDisallowInactiveBranch60Days()



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_negative_value_check import BaseResourceNegativeValueCheck


class APIManagementPublicAccess(BaseResourceNegativeValueCheck):
    def __init__(self) -> None:
        name = "Ensure API management public access is disabled"
        id = "CKV_AZURE_174"
        supported_resources = ("azurerm_api_management",)
        categories = (CheckCategories.NETWORKING,)
        super().__init__(
            name=name,
            id=id,
            categories=categories,
            supported_resources=supported_resources,
            missing_attribute_result=CheckResult.FAILED,
        )

    def get_inspected_key(self) -> str:
        return "public_network_access_enabled"

    def get_forbidden_values(self) -> list[Any]:
        return [True]


check = APIManagementPublicAccess()



from __future__ import annotations

from typing import Any, Dict, List, cast

from checkov.common.checks.base_check import BaseCheck
from checkov.common.checks.base_check_registry import BaseCheckRegistry
from checkov.common.models.enums import CheckResult
from checkov.common.typing import _SkippedCheck, _ScannerCallableAlias
from checkov.runner_filter import RunnerFilter
from checkov.yaml_doc.enums import BlockType
import jmespath

STARTLINE_MARK = "__startline__"

ENDLINE_MARK = "__endline__"


class Registry(BaseCheckRegistry):
    def __init__(self, report_type: str) -> None:
        super().__init__(report_type=report_type)
        self._scanner: dict[str, _ScannerCallableAlias] = {
            BlockType.ARRAY: self._scan_yaml_array,
            BlockType.OBJECT: self._scan_yaml_object,
        }

    def _scan_yaml_array(
            self, scanned_file: str, check: BaseCheck, skip_info: _SkippedCheck, entity: Dict[str, Any],
            entity_name: str,
            entity_type: str, results: Dict[str, Any]
    ) -> None:
        if isinstance(entity, dict):
            analyzed_entities = jmespath.search(entity_type, entity)
            if isinstance(analyzed_entities, dict):
                for item, item_conf in analyzed_entities.items():
                    if STARTLINE_MARK != item and ENDLINE_MARK != item:
                        self.update_result(
                            check=check,
                            entity_configuration=cast("dict[str, Any]", item_conf),
                            entity_name=item,
                            entity_type=entity_type,
                            results=results,
                            scanned_file=scanned_file,
                            skip_info=skip_info,
                        )
            if isinstance(analyzed_entities, list):
                for item in analyzed_entities:
                    if isinstance(item, str):
                        item = self.set_lines_for_item(item)
                    if STARTLINE_MARK != item and ENDLINE_MARK != item:
                        self.update_result(
                            check,
                            item,
                            entity_type,
                            entity_type,
                            results,
                            scanned_file,
                            skip_info
                        )
        if isinstance(entity, list):
            analyzed_entities = jmespath.search(entity_type, entity)
            if isinstance(analyzed_entities, list):
                for item in analyzed_entities:
                    if isinstance(item, str):
                        item = self.set_lines_for_item(item)
                    if STARTLINE_MARK != item and ENDLINE_MARK != item:
                        self.update_result(
                            check,
                            item,
                            entity_type,
                            entity_type,
                            results,
                            scanned_file,
                            skip_info
                        )
            else:
                for item in entity:
                    if entity_name in item:
                        result = self.update_result(
                            check,
                            item[entity_name],
                            entity_name,
                            entity_type,
                            results,
                            scanned_file,
                            skip_info
                        )
                        if result == CheckResult.FAILED:
                            break

    def _scan_yaml_object(
            self, scanned_file: str, check: BaseCheck, skip_info: _SkippedCheck, entity: Dict[str, Any],
            entity_name: str,
            entity_type: str, results: Dict[str, Any]
    ) -> None:
        if entity_name in entity:
            self.update_result(
                check,
                entity[entity_name],
                entity_name,
                entity_type,
                results,
                scanned_file,
                skip_info
            )

    def _scan_yaml_document(
            self, scanned_file: str, check: BaseCheck, skip_info: _SkippedCheck, entity: Dict[str, Any],
            entity_name: str,
            entity_type: str, results: Dict[str, Any]
    ) -> None:
        self.update_result(
            check, entity, entity_name, entity_type, results, scanned_file, skip_info
        )

    def _scan_yaml(
            self,
            scanned_file: str,
            checks: List[BaseCheck],
            skipped_checks: List[_SkippedCheck],
            runner_filter: RunnerFilter,
            entity: Dict[str, Any],
            entity_name: str,
            entity_type: str,
            results: Dict[str, Any],
    ) -> None:
        for check in checks:
            skip_info = ([x for x in skipped_checks if (x["id"] == check.id and entity[STARTLINE_MARK] <= x['line_number'] <= entity[ENDLINE_MARK])] or [{}])[0]

            if runner_filter.should_run_check(check=check, report_type=self.report_type):
                scanner = self._scanner.get(check.block_type, self._scan_yaml_document)
                if check.path:
                    target = entity
                    for p in check.path.split("."):
                        if p.endswith("]"):
                            ip = p.split("[")
                            i = int(ip[1][:-1])
                            target = target[ip[0]][i]
                        else:
                            target = target[p]
                else:
                    target = entity

                scanner(
                    scanned_file,
                    check,
                    skip_info,
                    target,
                    entity_name,
                    entity_type,
                    results,
                )

    def scan(  # type:ignore[override]  # return type is different than the base class
            self,
            scanned_file: str,
            entity: Dict[str, Any],
            skipped_checks: List[_SkippedCheck],
            runner_filter: RunnerFilter
    ) -> Dict[str, Any]:
        results: Dict[str, Any] = {}

        if not entity:
            return results

        for instruction, checks in self.checks.items():
            self._scan_yaml(
                scanned_file=scanned_file,
                checks=checks,
                skipped_checks=skipped_checks,
                runner_filter=runner_filter,
                entity=entity,
                entity_name=instruction,
                entity_type=instruction,
                results=results,
            )

        if self.wildcard_checks:
            for wildcard_pattern, checks in self.wildcard_checks.items():
                self._scan_yaml(
                    scanned_file=scanned_file,
                    checks=checks,
                    skipped_checks=skipped_checks,
                    runner_filter=runner_filter,
                    entity=entity,
                    entity_name=scanned_file,
                    entity_type=wildcard_pattern,
                    results=results,
                )

        return results

    def update_result(
            self,
            check: BaseCheck,
            entity_configuration: dict[str, Any],
            entity_name: str,
            entity_type: str,
            results: Dict[str, Any],
            scanned_file: str,
            skip_info: _SkippedCheck
    ) -> CheckResult:
        check_result = self.run_check(
            check,
            entity_configuration,
            entity_name,
            entity_type,
            scanned_file,
            skip_info,
        )
        result_key = self.get_result_key(check,
                                         entity_configuration,
                                         entity_name,
                                         entity_type,
                                         scanned_file,
                                         skip_info)

        result = check_result["result"]

        if isinstance(result, CheckResult) and result == CheckResult.SKIPPED:
            results[result_key] = {
                "check": check,
                "result": result,
                "suppress_comment": check_result["suppress_comment"],
                "results_configuration": None,
            }
            return result

        if isinstance(result, tuple):
            results[result_key] = {
                "check": check,
                "result": result[0],
                "results_configuration": result[1],
            }
            return result[0]
        results[result_key] = {
            "check": check,
            "result": result,
            "results_configuration": entity_configuration,
        }
        return result

    def get_result_key(self, check: BaseCheck,
                       entity_configuration: dict[str, Any],
                       entity_name: str,
                       entity_type: str,
                       scanned_file: str,
                       skip_info: _SkippedCheck) -> str:
        if isinstance(entity_configuration, dict) and STARTLINE_MARK in entity_configuration and ENDLINE_MARK in entity_configuration:
            return f'{entity_type}.{entity_name}.{check.id}[{entity_configuration[STARTLINE_MARK]}:{entity_configuration[ENDLINE_MARK]}]'

        if isinstance(entity_configuration, list):
            start_line = None
            end_line = None
            for sub_conf in entity_configuration:
                if STARTLINE_MARK in sub_conf and ENDLINE_MARK in sub_conf:
                    subconf_startline = sub_conf[STARTLINE_MARK]
                    sub_conf_endline = sub_conf[ENDLINE_MARK]
                    if not start_line:
                        start_line = subconf_startline
                    if not end_line:
                        end_line = sub_conf_endline
                    if sub_conf_endline > end_line:
                        end_line = sub_conf_endline
                    if subconf_startline < start_line:
                        start_line = subconf_startline
            if start_line and end_line:
                return f'{entity_type}.{entity_name}.{check.id}[{start_line}:{end_line}]'

        return f'{entity_type}.{entity_name}.{check.id}'

    def extract_entity_details(self, entity: dict[str, Any]) -> tuple[str, str, dict[str, Any]]:
        # not used, but is an abstractmethod
        return "", "", {}

    def set_lines_for_item(self, item: str) -> dict[int | str, str | int] | str:
        if not self.definitions_raw:
            return item

        item_lines = item.rstrip().split("\n")
        item_dict: dict[int | str, str | int] = {
            idx: line for idx, line in enumerate(item_lines)
        }

        if len(item_lines) == 1:
            item_line = item_lines[0]
            for idx, line in self.definitions_raw:
                if item_line in line:
                    item_dict[STARTLINE_MARK] = idx
                    item_dict[ENDLINE_MARK] = idx
            return item_dict

        first_line, last_line = item_lines[0], item_lines[-1]
        for idx, line in self.definitions_raw:
            if first_line in line:
                item_dict[STARTLINE_MARK] = idx
                continue

            if last_line in line:
                item_dict[ENDLINE_MARK] = idx
                break

        return item_dict



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.checks.resource.base_spec_check import BaseK8Check


class ShareHostIPC(BaseK8Check):
    def __init__(self) -> None:
        # CIS-1.3 1.7.3
        # CIS-1.5 5.2.3
        name = "Containers should not share the host IPC namespace"
        id = "CKV_K8S_18"
        # Location: Pod.spec.hostIPC
        # Location: CronJob.spec.jobTemplate.spec.template.spec.hostIPC
        # Location: *..spec.template.spec.hostIPC
        supported_kind = (
            "Pod",
            "Deployment",
            "DaemonSet",
            "StatefulSet",
            "ReplicaSet",
            "ReplicationController",
            "Job",
            "CronJob",
        )
        categories = (CheckCategories.KUBERNETES,)
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_kind)

    def scan_spec_conf(self, conf: dict[str, Any]) -> CheckResult:
        spec = {}

        if conf["kind"] == "Pod":
            if "spec" in conf:
                spec = conf["spec"]
        elif conf["kind"] == "CronJob":
            spec = conf.get("spec")
            if spec:
                job_template = spec.get("jobTemplate")
                if job_template:
                    job_template_spec = job_template.get("spec")
                    if job_template_spec:
                        template = job_template_spec.get("template")
                        if template:
                            if "spec" in template:
                                spec = template["spec"]
        else:
            inner_spec = self.get_inner_entry(conf, "spec")
            spec = inner_spec if inner_spec else spec
        if spec:
            if "hostIPC" in spec:
                if spec["hostIPC"]:
                    return CheckResult.FAILED
                else:
                    return CheckResult.PASSED
            return CheckResult.PASSED
        return CheckResult.FAILED


check = ShareHostIPC()



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.common.util.type_forcers import force_list
from checkov.common.util.type_forcers import force_int


class AbsSecurityGroupUnrestrictedIngress(BaseResourceCheck):
    def __init__(self, check_id: str, port: int) -> None:
        name = f"Ensure no security groups allow ingress from 0.0.0.0:0 to port {port}"
        supported_resources = ('aws_security_group', 'aws_security_group_rule')
        categories = (CheckCategories.NETWORKING,)
        super().__init__(name=name, id=check_id, categories=categories, supported_resources=supported_resources)
        self.port = port

    def scan_resource_conf(self, conf: dict[str, list[Any]]) -> CheckResult:
        """
            Looks for configuration at security group ingress rules :
            https://www.terraform.io/docs/providers/aws/r/security_group.html
            https://www.terraform.io/docs/providers/aws/r/security_group_rule.html

            Return PASS if:
            - The resource is an aws_security_group that contains no violating ingress rules (including if there are no
              ingress rules at all), OR
            - The resource is an aws_security_group_rule of type 'ingress' that does not violate the check.

            Return FAIL if:
            - The resource is an aws_security_group that contains a violating ingress rule, OR
            - The resource is an aws_security_group_rule of type 'ingress' that violates the check.

            Return UNKNOWN if:
            - the resource is an aws_security_group_rule of type 'egress', OR

        :param conf: aws_security_group configuration
        :return: <CheckResult>
        """

        if 'ingress' in conf:  # This means it's an SG resource with ingress block(s)
            ingress_conf = conf['ingress']
            for ingress_rule in ingress_conf:
                for rule in force_list(ingress_rule):
                    if isinstance(rule, dict) and self.contains_violation(rule):
                        self.evaluated_keys = [
                            f'ingress/[{ingress_conf.index(ingress_rule)}]/from_port',
                            f'ingress/[{ingress_conf.index(ingress_rule)}]/to_port',
                            f'ingress/[{ingress_conf.index(ingress_rule)}]/cidr_blocks',
                            f'ingress/[{ingress_conf.index(ingress_rule)}]/ipv6_cidr_blocks',
                        ]
                        return CheckResult.FAILED

            return CheckResult.PASSED

        if 'type' in conf:  # This means it's an SG_rule resource.
            type = force_list(conf['type'])[0]
            if type == 'ingress':
                self.evaluated_keys = ['from_port', 'to_port', 'cidr_blocks', 'ipv6_cidr_blocks']
                if self.contains_violation(conf):
                    return CheckResult.FAILED
                return CheckResult.PASSED
            return CheckResult.UNKNOWN

        # The result for an SG with no ingress block
        return CheckResult.PASSED

    def contains_violation(self, conf: dict[str, list[Any]]) -> bool:
        from_port = force_int(force_list(conf.get('from_port', [{-1}]))[0])
        to_port = force_int(force_list(conf.get('to_port', [{-1}]))[0])

        if from_port == 0 and to_port == 0:
            to_port = 65535

        if from_port is not None and to_port is not None and (from_port <= self.port <= to_port):
            conf_cidr_blocks = conf.get('cidr_blocks', [[]])
            if conf_cidr_blocks and len(conf_cidr_blocks) > 0:
                conf_cidr_blocks = conf_cidr_blocks[0]
            cidr_blocks = force_list(conf_cidr_blocks)
            if "0.0.0.0/0" in cidr_blocks:
                return True
            ipv6_cidr_blocks = conf.get('ipv6_cidr_blocks', [])
            if ipv6_cidr_blocks and ipv6_cidr_blocks[0] is not None and \
                    any(ip in ['::/0', '0000:0000:0000:0000:0000:0000:0000:0000/0'] for ip in ipv6_cidr_blocks[0]):
                return True
        return False



from __future__ import annotations

from abc import abstractmethod
from typing import Any

from checkov.common.images.image_referencer import Image
from checkov.common.util.consts import START_LINE, END_LINE


class WorkflowImageReferencerProvider:
    def __init__(self, workflow_config: dict[str, Any], file_path: str):
        self.workflow_config = workflow_config
        self.file_path = file_path

    @staticmethod
    def _get_start_end_lines(entity: dict[str, Any]) -> tuple[int, int]:
        return entity.get(START_LINE, 0), entity.get(END_LINE, 0)

    @abstractmethod
    def extract_images_from_workflow(self) -> list[Image]:
        pass



import os
import re
from dataclasses import dataclass
from pathlib import Path

from checkov.common.goget.github.get_git import GitGetter
from checkov.terraform.module_loading.content import ModuleContent

from checkov.terraform.module_loading.loader import ModuleLoader
from checkov.terraform.module_loading.module_params import ModuleParams

DEFAULT_MODULE_SOURCE_PREFIX = "git::https://"


@dataclass(frozen=True)
class ModuleSource:
    protocol: str
    root_module: str
    inner_module: str
    version: str
    username: str


class GenericGitLoader(ModuleLoader):
    def __init__(self):
        super().__init__()
        self.module_source_prefix = DEFAULT_MODULE_SOURCE_PREFIX

    @property
    def module_source_prefix(self):
        return self._module_source_prefix

    @module_source_prefix.setter
    def module_source_prefix(self, prefix):
        self._module_source_prefix = prefix

    def discover(self, module_params: ModuleParams):
        module_params.vcs_base_url = os.getenv("VCS_BASE_URL", "")  # format - https://example.com
        module_params.module_source_prefix = f"git::{module_params.vcs_base_url}" if module_params.vcs_base_url else None
        module_params.username = os.getenv("VCS_USERNAME", None)
        module_params.token = os.getenv("VCS_TOKEN", None)

    def _is_matching_loader(self, module_params: ModuleParams) -> bool:
        module_source_prefix = module_params.module_source_prefix if module_params.module_source_prefix else self.module_source_prefix
        if module_params.module_source.startswith(module_source_prefix):
            source = module_params.module_source.split(DEFAULT_MODULE_SOURCE_PREFIX)[-1]
            if module_params.token and module_params.username:
                module_params.module_source = f"{DEFAULT_MODULE_SOURCE_PREFIX}{module_params.username}:{module_params.token}@{source}"
            else:
                module_params.module_source = f"{DEFAULT_MODULE_SOURCE_PREFIX}{source}"
            return True
        # https://www.terraform.io/docs/modules/sources.html#generic-git-repository
        return module_params.module_source.startswith("git::")

    def _load_module(self, module_params: ModuleParams) -> ModuleContent:
        try:
            self._process_generic_git_repo(module_params)

            module_source = module_params.module_source.replace("git::", "")
            git_getter = GitGetter(module_source, create_clone_and_result_dirs=False)
            git_getter.temp_dir = module_params.dest_dir
            git_getter.do_get()
        except Exception as e:
            str_e = str(e)
            if 'File exists' not in str_e and 'already exists and is not an empty directory' not in str_e:
                self.logger.error(f"failed to get {module_params.module_source} because of {e}")
                return ModuleContent(dir=None, failed_url=module_params.module_source)
        return_dir = module_params.dest_dir
        if module_params.inner_module:
            return_dir = os.path.join(module_params.dest_dir, module_params.inner_module)
        return ModuleContent(dir=return_dir)

    def _find_module_path(self, module_params: ModuleParams) -> str:
        module_source = self._parse_module_source(module_params)
        module_path = Path(module_params.root_dir).joinpath(
            module_params.external_modules_folder_name,
            module_source.root_module,
            module_source.version,
            module_source.inner_module,
        )

        if module_params.inner_module:
            module_path = module_path / module_params.inner_module

        return str(module_path)

    def _parse_module_source(self, module_params: ModuleParams) -> ModuleSource:
        module_source_components = module_params.module_source.split("//")

        if "?ref=" in module_source_components[-1]:
            module_version_components = module_source_components[-1].rsplit("?ref=", maxsplit=1)
            module_source_components[-1] = module_version_components[0]
            version = module_version_components[1]
        else:
            version = "HEAD"

        if len(module_source_components) < 3:
            root_module = module_source_components[-1]
            inner_module = ""
        elif len(module_source_components) == 3:
            root_module = module_source_components[1]
            inner_module = module_source_components[2]
        else:
            raise Exception("invalid git url")

        username = re.match(re.compile(r"^(.*?@).*"), root_module)
        if username and username[1] != "git@":
            root_module = root_module.replace(username[1], "")

        if root_module.endswith(".git"):
            root_module = root_module[:-4]

        return ModuleSource(
            protocol=module_source_components[0], root_module=root_module, inner_module=inner_module, version=version,
            username=username[1] if username and username[1] != "git@" else ""
        )

    def _process_generic_git_repo(self, module_params: ModuleParams) -> None:
        module_source = self._parse_module_source(module_params)

        if module_source.inner_module:
            module_params.dest_dir = str(
                Path(module_params.root_dir).joinpath(
                    module_params.external_modules_folder_name, module_source.root_module, module_source.version
                )
            )
            module_params.inner_module = module_source.inner_module
            module_params.module_source = f"{module_source.protocol}//{module_source.root_module}"
            if module_source.username:
                module_params.module_source = f"{module_source.protocol}//{module_source.username}{module_source.root_module}"
            if module_source.version != "HEAD":
                module_params.module_source += f"?ref={module_source.version}"
        else:
            module_params.dest_dir = str(
                Path(module_params.root_dir).joinpath(
                    module_params.external_modules_folder_name, module_source.root_module, module_source.version
                )
            )


loader = GenericGitLoader()



from __future__ import annotations

import json
import re
from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Optional, Tuple

import hcl2


_FUNCTION_NAME_CHARS = frozenset("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789")

_ARG_VAR_PATTERN = re.compile(r"[a-zA-Z_]+(\.[a-zA-Z_]+)+")

TERRAFORM_NESTED_MODULE_PATH_PREFIX = '([{'
TERRAFORM_NESTED_MODULE_PATH_ENDING = '}])'
TERRAFORM_NESTED_MODULE_PATH_SEPARATOR_LENGTH = 3


@dataclass
class VarBlockMatch:
    full_str: str  # Example: ${local.foo}
    var_only: str  # Example: local.fop

    def replace(self, original: str, replaced: str) -> None:
        self.full_str = self.full_str.replace(original, replaced)
        self.var_only = self.var_only.replace(original, replaced)

    def is_simple_var(self) -> bool:
        """
        Indicates whether or not the value of the var block matches a "simple" var pattern. For example:
        local.blah, var.foo, module.one.a_resource.
        """
        return _ARG_VAR_PATTERN.match(self.var_only) is not None


class ParserMode(Enum):
    # Note: values are just for debugging.
    EVAL = "${"
    MAP = "{"
    STRING_SINGLE_QUOTE = "'"
    STRING_DOUBLE_QUOTE = '"'
    PARAMS = "("
    ARRAY = "["
    BLANK = " "

    @staticmethod
    def is_string(mode: "ParserMode") -> bool:
        return mode == ParserMode.STRING_SINGLE_QUOTE or mode == ParserMode.STRING_DOUBLE_QUOTE

    def __repr__(self) -> str:
        return str(self.value)

    def __str__(self) -> str:
        return str(self.value)


def find_var_blocks(value: str) -> List[VarBlockMatch]:
    """
    Find and return all the var blocks within a given string. Order is important and may contain portions of
    one another.
    """

    to_return: List[VarBlockMatch] = []

    mode_stack: List[ParserMode] = []
    eval_start_pos_stack: List[int] = []  # location of first char inside brackets
    param_start_pos_stack: List[int] = []  # location of open parens
    preceding_dollar = False
    preceding_string_escape = False
    # NOTE: function calls can be nested, but since param args are only being inspected for variables,
    #       it's alright to ignore outer calls.
    param_arg_start = -1
    for index, c in enumerate(value):
        current_mode = ParserMode.BLANK if not mode_stack else mode_stack[-1]

        # Print statement of power...
        # print(f"{str(index).ljust(3, ' ')} {c} {'$' if preceding_dollar else ' '} "
        #       f"{'`' if preceding_string_escape else ' '} "
        #       f"{current_mode.ljust(2)} - {mode_stack}")

        if c == "$":
            if preceding_dollar:  # ignore double $
                preceding_dollar = False
                continue

            preceding_dollar = True
            continue

        if c == "{" and preceding_dollar:
            mode_stack.append(ParserMode.EVAL)
            eval_start_pos_stack.append(index + 1)  # next char
            preceding_dollar = False
            continue
        elif c == "\\" and ParserMode.is_string(current_mode):
            preceding_string_escape = True
            continue

        preceding_dollar = False

        if c == "}":
            if current_mode == ParserMode.EVAL:
                mode_stack.pop()
                start_pos = eval_start_pos_stack.pop()
                eval_string = value[start_pos:index]
                to_return.append(VarBlockMatch("${" + eval_string + "}", eval_string))
            elif current_mode == ParserMode.MAP:
                mode_stack.pop()
        elif c == "]" and current_mode == ParserMode.ARRAY:
            mode_stack.pop()
        elif c == ")" and current_mode == ParserMode.PARAMS:
            if param_arg_start > 0:
                param_arg = value[param_arg_start:index].strip()
                if _ARG_VAR_PATTERN.match(param_arg):
                    to_return.append(VarBlockMatch(param_arg, param_arg))
                param_arg_start = -1

            mode_stack.pop()
            start_pos = param_start_pos_stack.pop()
            # See if these params are for a function call. Back up from the index to determine if there's
            # a function preceding.
            function_name_start_index = start_pos
            for function_index in range(start_pos - 1, 0, -1):
                if value[function_index] in _FUNCTION_NAME_CHARS:
                    function_name_start_index = function_index
                else:
                    break
            # We know now there's a function call here. But, don't call it out if it's directly wrapped
            # in eval markers.
            in_eval_markers = False
            if function_name_start_index >= 2:
                in_eval_markers = (
                    value[function_name_start_index - 2] == "$" and value[function_name_start_index - 1] == "{"
                )
            if function_name_start_index < start_pos and not in_eval_markers:
                to_return.append(
                    VarBlockMatch(
                        value[function_name_start_index : index + 1], value[function_name_start_index : index + 1]
                    )
                )
        elif c == '"':
            if preceding_string_escape:
                preceding_string_escape = False
                continue
            elif current_mode == ParserMode.STRING_DOUBLE_QUOTE:
                mode_stack.pop()
            else:
                mode_stack.append(ParserMode.STRING_DOUBLE_QUOTE)
        elif c == "'":
            if preceding_string_escape:
                preceding_string_escape = False
                continue
            elif current_mode == ParserMode.STRING_SINGLE_QUOTE:
                mode_stack.pop()
            else:
                mode_stack.append(ParserMode.STRING_SINGLE_QUOTE)
        elif c == "{":
            # NOTE: Can't be preceded by a dollar sign (that was checked earlier)
            if not ParserMode.is_string(current_mode):
                mode_stack.append(ParserMode.MAP)
        elif c == "[":  # do we care?
            if not ParserMode.is_string(current_mode):
                mode_stack.append(ParserMode.ARRAY)
        elif c == "(":  # do we care?
            if not ParserMode.is_string(current_mode):
                mode_stack.append(ParserMode.PARAMS)
                param_start_pos_stack.append(index)
                param_arg_start = index + 1
        elif c == ",":
            if current_mode == ParserMode.PARAMS and param_arg_start > 0:
                param_arg = value[param_arg_start:index].strip()
                if _ARG_VAR_PATTERN.match(param_arg):
                    to_return.append(VarBlockMatch(param_arg, param_arg))
                param_arg_start = index + 1
        elif c == "?" and current_mode == ParserMode.EVAL:  # ternary
            # If what's been processed in the ternary so far is "true" or "false" (boolean or string type)
            # then nothing special will happen here and only the full expression will be returned.
            # Anything else will be treated as an unresolved variable block.
            start_pos = eval_start_pos_stack[-1]  # DO NOT pop: there's no separate eval start indicator
            eval_string = value[start_pos:index].strip()

            # HACK ALERT: For the cases with the trailing quotes, see:
            #             test_hcl2_load_assumptions.py -> test_weird_ternary_string_clipping
            if eval_string not in {"true", "false", '"true"', '"false"', 'true"', 'false"'}:
                # REMINDER: The eval string is not wrapped in a eval markers since they didn't really
                #           appear in the original value. If they're put in, substitution doesn't
                #           work properly.
                to_return.append(VarBlockMatch(eval_string, eval_string))

        preceding_string_escape = False

    return to_return


def split_merge_args(value: str) -> list[str] | None:
    """
    Split arguments of a merge function. For example, "merge(local.one, local.two)" would
    call this function with a value of "local.one, local.two" which would return
    ["local.one", "local.two"]. If the value cannot be unpacked, None will be returned.
    """
    if not value:
        return None

    # There are a number of splitting scenarios depending on whether variables or
    # direct maps are used:
    #           merge({tag1="foo"},{tag2="bar"})
    #           merge({tag1="foo"},local.some_tags)
    #           merge(local.some_tags,{tag2="bar"})
    #           merge(local.some_tags,local.some_other_tags)
    # Also, the number of arguments can vary, things can be nested, strings are evil...
    # See tests/terraform/test_parser_var_blocks.py for many examples.

    to_return = []
    current_arg_buffer = ""
    processing_str_escape = False
    inside_collection_stack: List[str] = []  # newest at position 0, contains the terminator for the collection
    for c in value:
        if c == "," and not inside_collection_stack:
            current_arg_buffer = current_arg_buffer.strip()
            # Note: can get a zero-length buffer when there's a double comman. This can
            #       happen with multi-line args (see parser_internals test)
            if len(current_arg_buffer) != 0:
                to_return.append(current_arg_buffer)
            current_arg_buffer = ""
        else:
            current_arg_buffer += c

        processing_str_escape = _str_parser_loop_collection_helper(c, inside_collection_stack, processing_str_escape)

    current_arg_buffer = current_arg_buffer.strip()
    if len(current_arg_buffer) > 0:
        to_return.append(current_arg_buffer)

    if len(to_return) == 0:
        return None
    return to_return


def _str_parser_loop_collection_helper(c: str, inside_collection_stack: List[str], processing_str_escape: bool) -> bool:
    """
    This function handles dealing with tracking when a char-by-char state loop is inside a
    "collection" (map, array index, method args, string).

    :param c:       Active character
    :param inside_collection_stack:     Stack of terminators for collections. This will be modified by
                                        this function. The active terminator will be at position 0.


    :return: value to set for `processing_str_escape`
    """
    inside_a_string = False
    if inside_collection_stack:
        terminator = inside_collection_stack[0]

        if terminator == '"' or terminator == "'":
            if processing_str_escape:
                processing_str_escape = False
                return processing_str_escape
            elif c == "\\":
                processing_str_escape = True
                return processing_str_escape
            else:
                inside_a_string = True

        if c == terminator:
            del inside_collection_stack[0]
            return processing_str_escape

    if not inside_a_string:
        if c == '"':
            inside_collection_stack.insert(0, '"')
        elif c == "'":
            inside_collection_stack.insert(0, "'")
        elif c == "{":
            inside_collection_stack.insert(0, "}")
        elif c == "[":
            inside_collection_stack.insert(0, "]")
        elif c == "(":
            inside_collection_stack.insert(0, ")")

    return processing_str_escape


def eval_string(value: str) -> Any:
    try:
        value_string = value.replace("'", '"')
        parsed = hcl2.loads(f"eval = {value_string}\n")  # NOTE: newline is needed
        return parsed["eval"][0]
    except Exception:
        return None


def string_to_native(value: str) -> Any:
    try:
        value_string = value.replace("'", '"')
        return json.loads(value_string)
    except Exception:
        return None


def to_string(value: Any) -> str:
    if value is True:
        return "true"
    elif value is False:
        return "false"
    return str(value)


def get_current_module_index(full_path: str) -> Optional[int]:
    hcl_index = None
    tf_index = None
    if TERRAFORM_NESTED_MODULE_PATH_PREFIX not in full_path and '#' not in full_path:
        return len(full_path)
    if f'.hcl{TERRAFORM_NESTED_MODULE_PATH_PREFIX}' in full_path:
        hcl_index = full_path.index(f'.hcl{TERRAFORM_NESTED_MODULE_PATH_PREFIX}') + 4  # len('.hcl')
    elif '.hcl#' in full_path:
        hcl_index = full_path.index('.hcl#') + 4  # len('.hcl')
    if f'.tf{TERRAFORM_NESTED_MODULE_PATH_PREFIX}' in full_path:
        tf_index = full_path.index(f'.tf{TERRAFORM_NESTED_MODULE_PATH_PREFIX}') + 3    # len('.tf')
    elif '.tf#' in full_path:
        tf_index = full_path.index('.tf#') + 3  # len('.tf')
    if hcl_index and tf_index:
        # returning the index of the first file
        return min(hcl_index, tf_index)
    if hcl_index:
        return hcl_index
    return tf_index


def is_nested(full_path: str) -> bool:
    return TERRAFORM_NESTED_MODULE_PATH_PREFIX in full_path


def get_tf_definition_key(nested_module: str, module_name: str, module_index: Any, nested_key: str = '') -> str:
    return f"{nested_module}{TERRAFORM_NESTED_MODULE_PATH_PREFIX}{module_name}#{module_index}{nested_key}{TERRAFORM_NESTED_MODULE_PATH_ENDING}"


def get_tf_definition_key_from_module_dependency(path: str, module_dependency: str, module_dependency_num: str) -> str:
    if not module_dependency:
        return path
    if not is_nested(module_dependency):
        return f"{path}{TERRAFORM_NESTED_MODULE_PATH_PREFIX}{module_dependency}#{module_dependency_num}{TERRAFORM_NESTED_MODULE_PATH_ENDING}"
    module_index = get_current_module_index(module_dependency)
    return f"{path}{TERRAFORM_NESTED_MODULE_PATH_PREFIX}{module_dependency[:module_index]}#{module_dependency_num}{module_dependency[module_index:]}{TERRAFORM_NESTED_MODULE_PATH_ENDING}"


def get_module_from_full_path(file_path: str) -> Tuple[Optional[str], Optional[str]]:
    if not is_nested(file_path):
        return None, None
    tmp_path = file_path[file_path.index(TERRAFORM_NESTED_MODULE_PATH_PREFIX) + TERRAFORM_NESTED_MODULE_PATH_SEPARATOR_LENGTH: -TERRAFORM_NESTED_MODULE_PATH_SEPARATOR_LENGTH]
    if is_nested(tmp_path):
        module = get_abs_path(tmp_path) + tmp_path[tmp_path.index(TERRAFORM_NESTED_MODULE_PATH_PREFIX):]
        index = tmp_path[tmp_path.index('#') + 1:tmp_path.index(TERRAFORM_NESTED_MODULE_PATH_PREFIX)]
    else:
        module = get_abs_path(tmp_path)
        index = tmp_path[tmp_path.index('#') + 1:]
    return module, index


def get_abs_path(file_path: str) -> str:
    return file_path[:get_current_module_index(file_path)]



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.common.util.type_forcers import force_list
from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceCheck


class GKEMetadataServerIsEnabled(BaseResourceCheck):
    def __init__(self):
        name = "Ensure the GKE Metadata Server is Enabled"
        id = "CKV_GCP_69"
        supported_resources = ['google_container_cluster', 'google_container_node_pool']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        """
        looks for workload_metadata_config in node_config which ensures that the metadata server is enabled
        https://registry.terraform.io/providers/hashicorp/google/latest/docs/resources/container_cluster
        :param conf: google_container_cluster or google_container_node_pool configuration
        :return: <CheckResult>
        """
        if 'node_config' in conf.keys():
            node = conf["node_config"][0]
            if isinstance(node, dict) and 'workload_metadata_config' in node:
                workload_metadatas = force_list(node["workload_metadata_config"][0])
                for workload_metadata in workload_metadatas:
                    if workload_metadata.get("mode", None) == ["GKE_METADATA"]:
                        return CheckResult.PASSED
                    # deprecated in newer Google provider
                    elif workload_metadata.get("node_metadata", None) == ["GKE_METADATA_SERVER"]:
                        return CheckResult.PASSED

        return CheckResult.FAILED


check = GKEMetadataServerIsEnabled()



from typing import Any

from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceValueCheck
from checkov.common.models.enums import CheckCategories
from checkov.common.models.consts import ANY_VALUE


class K8SSecurityGroup(BaseResourceValueCheck):
    def __init__(self) -> None:
        name = "Ensure security group is assigned to Kubernetes cluster."
        id = "CKV_YC_14"
        supported_resources = ("yandex_kubernetes_cluster",)
        categories = (CheckCategories.NETWORKING,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self) -> str:
        return "master/[0]/security_group_ids"

    def get_expected_value(self) -> Any:
        return ANY_VALUE


check = K8SSecurityGroup()



from enum import Enum


class GraphSource(str, Enum):
    ANSIBLE = "Ansible"
    BICEP = "Bicep"
    CLOUDFORMATION = "CloudFormation"
    DOCKERFILE = "Dockerfile"
    GITHUB_ACTIONS = "GitHubActions"
    KUBERNETES = "Kubernetes"
    TERRAFORM = "Terraform"

    def __str__(self) -> str:
        # needed, because of a Python 3.11 change
        return self.value



import ast


def get_model_parts_info(model_ast):
    parts_info = []
    for child_node in model_ast.body:
        node_type = get_model_node_type(child_node)
        parts_info.append({
            'model_name': model_ast.name,
            'node': child_node,
            'type': node_type
        })
    return parts_info


def get_model_node_type(child_node) -> str:
    direct_node_types_mapping = [
        (ast.If, lambda n: 'if'),
        (ast.Pass, lambda n: 'pass'),
        ((ast.Assign, ast.AnnAssign), lambda n: get_assighment_type(n)),
        ((ast.FunctionDef, ast.AsyncFunctionDef), lambda n: get_funcdef_type(n)),
        (ast.Expr, lambda n: 'docstring' if isinstance(n.value, ast.Str) else 'expression'),
        (ast.ClassDef, lambda n: 'meta_class' if child_node.name == 'Meta' else 'nested_class'),
    ]
    for type_or_type_tuple, type_getter in direct_node_types_mapping:
        if isinstance(child_node, type_or_type_tuple):  # type: ignore
            return type_getter(child_node)


def get_assighment_type(child_node) -> str:
    assignee_node = child_node.target if isinstance(child_node, ast.AnnAssign) else child_node.targets[0]
    assighment_type = 'field'
    if isinstance(assignee_node, ast.Subscript):
        assighment_type = 'expression'
    if isinstance(assignee_node, ast.Name) and is_caps_lock_str(assignee_node.id):
        assighment_type = 'constant'
    if isinstance(child_node.value, ast.Call):
        dump_callable = ast.dump(child_node.value.func)
        if (
            'ForeignKey' in dump_callable
            or 'ManyToManyField' in dump_callable
            or 'OneToOneField' in dump_callable
            or 'GenericRelation' in dump_callable
        ):
            assighment_type = 'outer_field'
    return assighment_type


def get_funcdef_type(child_node) -> str:
    special_methods_names = {
        '__new__',
        '__init__',
        '__post_init__',
        '__str__',
        'save',
        'delete',
    }
    decorator_names_to_types_map = {
        'property': 'property_method',
        'cached_property': 'property_method',
        'staticmethod': 'static_method',
        'classmethod': 'class_method',

        'private_property': 'private_property_method',
        'private_cached_property': 'private_property_method',
        'private_staticmethod': 'private_static_method',
        'private_classmethod': 'private_class_method',
    }
    for decorator_info in child_node.decorator_list:
        if (
            isinstance(decorator_info, ast.Name)
            and decorator_info.id in decorator_names_to_types_map
        ):

            if child_node.name.startswith('_'):
                return decorator_names_to_types_map[f'private_{decorator_info.id}']

            return decorator_names_to_types_map[decorator_info.id]
    funcdef_type = 'method'
    if child_node.name in special_methods_names:
        funcdef_type = child_node.name
    elif child_node.name.startswith('__') and child_node.name.endswith('__'):
        funcdef_type = 'magic_method'
    elif child_node.name.startswith('_'):
        funcdef_type = 'private_method'
    return funcdef_type


def is_caps_lock_str(var_name: str) -> bool:
    return var_name.upper() == var_name



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class DockerSocketVolume(BaseResourceCheck):
    def __init__(self) -> None:
        # Exposing the socket gives container information and increases risk of exploit
        # read-only is not a solution but only makes it harder to exploit.
        # Location: Pod.spec.volumes[].hostPath.path
        # Location: CronJob.spec.jobTemplate.spec.template.spec.volumes[].hostPath.path
        # Location: *.spec.template.spec.volumes[].hostPath.path
        id = "CKV_K8S_27"
        name = "Do not expose the docker daemon socket to containers"
        supported_resources = ("kubernetes_pod", "kubernetes_pod_v1",
                               "kubernetes_deployment", "kubernetes_deployment_v1",
                               "kubernetes_daemonset", "kubernetes_daemon_set_v1")
        categories = (CheckCategories.NETWORKING,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: dict[str, list[Any]]):
        if "spec" not in conf:
            self.evaluated_keys = [""]
            return CheckResult.FAILED

        spec = conf['spec'][0]
        if not spec:
            return CheckResult.UNKNOWN

        if "volume" in spec and spec.get("volume"):
            volumes = spec.get("volume")
            for idx, v in enumerate(volumes):
                if v.get("host_path"):
                    if "path" in v["host_path"][0]:
                        if v["host_path"][0]["path"] == ["/var/run/docker.sock"]:
                            self.evaluated_keys = [f"spec/volume/{idx}/host_path/[0]/path"]
                            return CheckResult.FAILED
        if "template" in spec and spec.get("template"):
            template = spec.get("template")[0]
            if "spec" in template:
                temp_spec = template.get("spec")[0]
                if "volume" in temp_spec and temp_spec.get("volume"):
                    volumes = temp_spec.get("volume")
                    for idx, v in enumerate(volumes):
                        if isinstance(v, dict) and v.get("host_path"):
                            if "path" in v["host_path"][0]:
                                path = v["host_path"][0]["path"]
                                if path == ["/var/run/docker.sock"]:
                                    self.evaluated_keys = [f"spec/template/spec/volume/{idx}/host_path/[0]/path"]
                                    return CheckResult.FAILED

        return CheckResult.PASSED


check = DockerSocketVolume()



from __future__ import annotations

import logging
from typing import Set, Any, Generator, Pattern, Optional, Dict, Tuple, TYPE_CHECKING, cast


from detect_secrets.constants import VerifiedResult
from detect_secrets.core.potential_secret import PotentialSecret
from detect_secrets.plugins.base import RegexBasedDetector
from detect_secrets.util.inject import call_function_with_arguments
import re

from checkov.secrets.plugins.load_detectors import load_detectors

MIN_CHARACTERS = 5
MAX_CHARACTERS = 100

if TYPE_CHECKING:
    from detect_secrets.util.code_snippet import CodeSnippet


class CustomRegexDetector(RegexBasedDetector):
    secret_type = "Regex Detector"  # noqa: CCE003 # nosec
    denylist: Set[Pattern[str]] = set()  # noqa: CCE003

    def __init__(self) -> None:
        self.regex_to_metadata: dict[str, dict[str, Any]] = dict()
        self.denylist = set()
        detectors = load_detectors()

        for detector in detectors:
            self.denylist.add(re.compile('{}'.format(detector["Regex"])))
            self.regex_to_metadata[detector["Regex"]] = detector

    def analyze_line(
            self,
            filename: str,
            line: str,
            line_number: int = 0,
            context: Optional[CodeSnippet] = None,
            raw_context: Optional[CodeSnippet] = None,
            **kwargs: Any
    ) -> Set[PotentialSecret]:
        """This examines a line and finds all possible secret values in it."""
        output: Set[PotentialSecret] = set()
        for match, regex in self.analyze_string(line, **kwargs):
            try:
                verified_result = call_function_with_arguments(self.verify, secret=match, context=context)
                is_verified = True if verified_result == VerifiedResult.VERIFIED_TRUE else False
            except Exception:
                is_verified = False
            regex_data = self.regex_to_metadata[regex.pattern]
            ps = PotentialSecret(type=regex_data["Name"], filename=filename, secret=match,
                                 line_number=line_number, is_verified=is_verified)
            ps.check_id = self.regex_to_metadata[regex.pattern]["Check_ID"]  # type:ignore[attr-defined]
            if len(cast(str, ps.secret_value)) in range(MIN_CHARACTERS, MAX_CHARACTERS) or not regex_data['isCustom']:
                output.add(ps)
            else:
                logging.info(
                    f'Finding for check {ps.check_id} are not 5-100 characters in length, was ignored')  # type: ignore

        return output

    def analyze_string(self, string: str, **kwargs: Optional[Dict[str, Any]]) -> Generator[Tuple[str, Pattern[str]], None, None]:  # type:ignore[override]
        for regex in self.denylist:
            for match in regex.findall(string):
                if isinstance(match, tuple):
                    for submatch in filter(bool, match):
                        # It might make sense to paste break after yielding
                        yield submatch, regex
                else:
                    yield match, regex



import os
from typing import List
from checkov.common.runners.base_runner import strtobool

PATH_SEPARATOR = "->"


def unify_dependency_path(dependency_path: List[str]) -> str:
    if not dependency_path:
        return ''
    if strtobool(os.getenv('CHECKOV_ENABLE_NESTED_MODULES', 'False')):
        return dependency_path[-1]
    return PATH_SEPARATOR.join(dependency_path)



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.cloudformation.checks.resource.base_resource_check import BaseResourceCheck
from checkov.common.util.type_forcers import force_list


class GlueSecurityConfiguration(BaseResourceCheck):
    def __init__(self) -> None:
        name = "Ensure Glue Security Configuration Encryption is enabled"
        id = "CKV_AWS_99"
        supported_resources = ('AWS::Glue::SecurityConfiguration',)
        categories = (CheckCategories.LOGGING,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: dict[str, Any]) -> CheckResult:
        s3_enc = False
        cw_enc = False
        book_enc = False
        if 'Properties' in conf.keys():
            if 'EncryptionConfiguration' in conf['Properties'].keys():
                enc_conf = conf['Properties']['EncryptionConfiguration']

                if 'CloudWatchEncryption' in enc_conf.keys():
                    if 'CloudWatchEncryptionMode' in enc_conf['CloudWatchEncryption'].keys():
                        if enc_conf['CloudWatchEncryption']['CloudWatchEncryptionMode'] != 'DISABLED':
                            cw_enc = True

                if 'JobBookmarksEncryption' in enc_conf.keys():
                    if 'JobBookmarksEncryptionMode' in enc_conf['JobBookmarksEncryption'].keys():
                        if enc_conf['JobBookmarksEncryption']['JobBookmarksEncryptionMode'] != 'DISABLED':
                            book_enc = True

                if 'S3Encryptions' in enc_conf.keys():
                    for s3_encryption in force_list(enc_conf['S3Encryptions']):
                        if 'S3EncryptionMode' in s3_encryption.keys():
                            if s3_encryption['S3EncryptionMode'] != 'DISABLED':
                                s3_enc = True
                                break

        if s3_enc and cw_enc and book_enc:
            return CheckResult.PASSED
        
        return CheckResult.FAILED


check = GlueSecurityConfiguration()



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Optional

from checkov.azure_pipelines.checks.registry import registry
from checkov.azure_pipelines.common.resource_id_utils import generate_resource_key_recursive
from checkov.azure_pipelines.image_referencer.manager import AzurePipelinesImageReferencerManager
from checkov.common.images.image_referencer import ImageReferencerMixin, Image
from checkov.common.output.report import CheckType, Report
from checkov.runner_filter import RunnerFilter
from checkov.yaml_doc.runner import Runner as YamlRunner


if TYPE_CHECKING:
    from checkov.common.checks.base_check_registry import BaseCheckRegistry
    from collections.abc import Iterable
    from networkx import DiGraph


class Runner(ImageReferencerMixin["dict[str, dict[str, Any] | list[dict[str, Any]]]"], YamlRunner):
    check_type = CheckType.AZURE_PIPELINES  # noqa: CCE003  # a static attribute

    def require_external_checks(self) -> bool:
        return False

    def import_registry(self) -> BaseCheckRegistry:
        return registry

    def _parse_file(
        self, f: str, file_content: str | None = None
    ) -> tuple[dict[str, Any] | list[dict[str, Any]], list[tuple[int, str]]] | None:
        if self.is_workflow_file(f):
            return super()._parse_file(f=f)
        return None

    def is_workflow_file(self, file_path: str) -> bool:
        return file_path.endswith(('azure-pipelines.yml', 'azure-pipelines.yaml'))

    def get_resource(self, file_path: str, key: str, supported_entities: Iterable[str],
                     start_line: int = -1, end_line: int = -1) -> str:
        if not self.definitions or not isinstance(self.definitions, dict):
            return key
        resource_name: Optional[str] = generate_resource_key_recursive(start_line, end_line, self.definitions[file_path])
        return resource_name if resource_name else key

    def run(
            self,
            root_folder: str | None = None,
            external_checks_dir: list[str] | None = None,
            files: list[str] | None = None,
            runner_filter: RunnerFilter | None = None,
            collect_skip_comments: bool = True,
    ) -> Report | list[Report]:
        runner_filter = runner_filter or RunnerFilter()
        report = super().run(root_folder=root_folder, external_checks_dir=external_checks_dir,
                             files=files, runner_filter=runner_filter, collect_skip_comments=collect_skip_comments)
        if runner_filter.run_image_referencer:
            if files:
                # 'root_folder' shouldn't be empty to remove the whole path later and only leave the shortened form
                root_folder = os.path.split(os.path.commonprefix(files))[0]

            image_report = self.check_container_image_references(
                root_path=root_folder,
                runner_filter=runner_filter,
                definitions=self.definitions
            )

            if image_report:
                return [report, image_report]  # type:ignore[list-item]  # report can only be of type Report, not a list

        return report

    def extract_images(
            self,
            graph_connector: DiGraph | None = None,
            definitions: dict[str, dict[str, Any] | list[dict[str, Any]]] | None = None,
            definitions_raw: dict[str, list[tuple[int, str]]] | None = None
    ) -> list[Image]:
        images: list[Image] = []
        if not definitions:
            return images

        for file, config in definitions.items():
            if isinstance(config, list):
                continue
            is_container = False
            for row in self.definitions_raw.get(file, []):
                if 'container' in row[1]:
                    is_container = True
                    break
            if not is_container:
                return images
            manager = AzurePipelinesImageReferencerManager(workflow_config=config, file_path=file)
            images.extend(manager.extract_images_from_workflow())

        return images



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Any

from checkov.common.bridgecrew.platform_integration import BcPlatformIntegration
from checkov.common.bridgecrew.vulnerability_scanning.integrations.twistcli import TwistcliIntegration
from checkov.common.util.str_utils import removeprefix


class PackageScanningIntegration(TwistcliIntegration):
    def create_report(
        self,
        twistcli_scan_result: dict[str, Any],
        bc_platform_integration: BcPlatformIntegration,
        file_path: Path,
        **kwargs: Any,
    ) -> dict[str, Any]:
        if not bc_platform_integration.bc_source:
            logging.error("Source was not set")
            return {}

        payload: dict[str, Any] = {
            "packageName": str(file_path.name),
            "packageFilePath": removeprefix(str(file_path), os.getenv("BC_ROOT_DIR", "")),
            "type": "Package",
            "sourceId": bc_platform_integration.repo_id,
            "branch": bc_platform_integration.repo_branch,
            "sourceType": bc_platform_integration.bc_source.name,
            "vulnerabilities": self.get_vulnerabilities_for_report(twistcli_scan_result),
            "packages": self.get_packages_for_report(twistcli_scan_result),
        }
        if bc_platform_integration.cicd_details:
            payload["cicdDetails"] = bc_platform_integration.cicd_details
        return payload


package_scanning_integration = PackageScanningIntegration()



from __future__ import annotations

import os
from abc import abstractmethod
from typing import TYPE_CHECKING, Callable, Any, Mapping, Union

from checkov.common.graph.graph_builder import CustomAttributes
from checkov.common.images.image_referencer import Image

if TYPE_CHECKING:
    import networkx
    import igraph
    from typing_extensions import TypeAlias

_ExtractImagesCallableAlias: TypeAlias = Callable[["dict[str, Any]"], "list[str]"]


class GraphImageReferencerProvider:
    __slots__ = ("graph_connector", "supported_resource_types", "graph_framework")

    # TODO add to graph_connector type fot igraph and implement the extract_nodes_igraph function
    def __init__(self, graph_connector: Union[igraph.Graph, networkx.DiGraph],
                 supported_resource_types: dict[str, _ExtractImagesCallableAlias] | Mapping[
                     str, _ExtractImagesCallableAlias]):
        self.graph_connector = graph_connector
        self.supported_resource_types = supported_resource_types
        self.graph_framework = os.environ.get('CHECKOV_GRAPH_FRAMEWORK', 'NETWORKX')

    @abstractmethod
    def extract_images_from_resources(self) -> list[Image]:
        pass

    def extract_nodes(self) -> networkx.Graph | igraph.Graph:
        if self.graph_framework == 'NETWORKX':
            return self.extract_nodes_networkx()
        else:
            return self.extract_nodes_igraph()

    def extract_nodes_networkx(self) -> networkx.Graph:
        resource_nodes = [
            node
            for node, resource_type in self.graph_connector.nodes(data=CustomAttributes.RESOURCE_TYPE)
            if resource_type and resource_type in self.supported_resource_types
        ]

        return self.graph_connector.subgraph(resource_nodes)

    def extract_nodes_igraph(self) -> igraph.Graph:
        pass



from __future__ import annotations

from typing import List, TYPE_CHECKING

from checkov.terraform.checks.data.BaseCloudsplainingIAMCheck import BaseCloudsplainingIAMCheck

if TYPE_CHECKING:
    from cloudsplaining.scan.policy_document import PolicyDocument


class CloudSplainingDataExfiltration(BaseCloudsplainingIAMCheck):
    def __init__(self) -> None:
        name = "Ensure IAM policies does not allow data exfiltration"
        id = "CKV_AWS_108"
        super().__init__(name=name, id=id)

    def cloudsplaining_analysis(self, policy: PolicyDocument) -> List[str]:
        return policy.allows_data_exfiltration_actions


check = CloudSplainingDataExfiltration()



from checkov.common.models.enums import CheckCategories
from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceValueCheck


class LaunchConfigurationEncryptionVPC(BaseResourceValueCheck):
    def __init__(self) -> None:
        name = "Ensure Basic Block storage is encrypted."
        id = "CKV_NCP_7"
        supported_resources = ("ncloud_launch_configuration",)
        categories = (CheckCategories.ENCRYPTION,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self) -> str:
        return "is_encrypted_volume"


check = LaunchConfigurationEncryptionVPC()



from __future__ import annotations

import re
from pathlib import Path
from typing import Any

from checkov.common.parsers.yaml.parser import parse

TASK_NAME_PATTERN = re.compile(r"^\s*-\s+name:\s+", re.MULTILINE)

# https://docs.ansible.com/ansible/latest/reference_appendices/playbooks_keywords.html#task
TASK_RESERVED_KEYWORDS = {
    "action",
    "any_errors_fatal",
    "args",
    "async",
    "become",
    "become_exe",
    "become_flags",
    "become_method",
    "become_user",
    "changed_when",
    "check_mode",
    "collections",
    "connection",
    "debugger",
    "delay",
    "delegate_facts",
    "delegate_to",
    "diff",
    "environment",
    "failed_when",
    "ignore_errors",
    "ignore_unreachable",
    "local_action",
    "loop",
    "loop_control",
    "module_defaults",
    "name",
    "no_log",
    "notify",
    "poll",
    "port",
    "register",
    "remote_user",
    "retries",
    "run_once",
    "tags",
    "throttle",
    "timeout",
    "until",
    "vars",
    "when",
}


def get_scannable_file_paths(root_folder: str | Path) -> set[Path]:
    """Finds yaml files"""

    file_paths: set[Path] = set()

    if root_folder:
        root_path = root_folder if isinstance(root_folder, Path) else Path(root_folder)
        file_paths = {file_path for file_path in root_path.rglob("*.[y][am]*[l]") if file_path.is_file()}

    return file_paths


def get_relevant_file_content(file_path: str | Path) -> str | None:
    if not str(file_path).endswith((".yaml", ".yml")):
        return None

    content = Path(file_path).read_text()
    match_task_name = re.search(TASK_NAME_PATTERN, content)
    if match_task_name:
        # there are more files, which belong to an ansible playbook,
        # but we are currently only interested in 'tasks'
        return content

    return None


def parse_file(
    f: str | Path, file_content: str | None = None
) -> tuple[dict[str, Any] | list[dict[str, Any]], list[tuple[int, str]]] | None:
    file_content = get_relevant_file_content(file_path=f)
    if file_content:
        content = parse(filename=str(f), file_content=file_content)
        return content

    return None



from __future__ import annotations

import logging
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import Any

from checkov.common.graph.graph_builder import CustomAttributes
from checkov.common.graph.graph_builder.consts import GraphSource
from checkov.common.graph.graph_builder.graph_components.block_types import BlockType
from checkov.common.graph.graph_builder.graph_components.blocks import Block
from checkov.common.runners.graph_builder.local_graph import ObjectLocalGraph
from checkov.common.util.consts import START_LINE, END_LINE
from checkov.github_actions.graph_builder.graph_components.resource_types import ResourceType
from checkov.github_actions.utils import get_scannable_file_paths, parse_file


class GitHubActionsLocalGraph(ObjectLocalGraph):
    def __init__(self, definitions: dict[str | Path, dict[str, Any] | list[dict[str, Any]]]) -> None:
        super().__init__(definitions=definitions)

        self.source = GraphSource.GITHUB_ACTIONS
        self.job_steps_map: "dict[tuple[str, str], list[tuple[str, str]]]" = defaultdict(list)

    def _create_vertices(self) -> None:
        for file_path, definition in self.definitions.items():
            if not isinstance(definition, dict):
                logging.debug(f"definition of file {file_path} has the wrong type {type(definition)}")
                return

            file_path = str(file_path)

            self._create_jobs_vertices(file_path=file_path, jobs=definition.get(ResourceType.JOBS))
            self._create_steps_vertices(file_path=file_path, jobs=definition.get(ResourceType.JOBS))
            self._create_permissions_vertices(file_path=file_path, permissions=definition.get(ResourceType.PERMISSIONS))
            self._create_on_vertices(file_path=file_path, on=definition.get(ResourceType.ON))

    def _create_jobs_vertices(self, file_path: str, jobs: Any) -> None:
        """Creates jobs vertices"""

        if not jobs or not isinstance(jobs, dict):
            return

        for name, config in jobs.items():
            if name in (START_LINE, END_LINE):
                continue

            attributes = deepcopy(config)
            attributes[CustomAttributes.RESOURCE_TYPE] = ResourceType.JOBS

            block_name = f"{ResourceType.JOBS}.{name}"

            self.vertices.append(
                Block(
                    name=block_name,
                    config=config,
                    path=file_path,
                    block_type=BlockType.RESOURCE,
                    attributes=attributes,
                    id=block_name,
                    source=self.source,
                )
            )

    def _create_steps_vertices(self, file_path: str, jobs: Any) -> None:
        """Creates steps vertices from jobs"""

        if not jobs or not isinstance(jobs, dict):
            return

        for name, job_config in jobs.items():
            if name in (START_LINE, END_LINE):
                continue

            steps = job_config.get(ResourceType.STEPS)

            if not steps or not isinstance(steps, list):
                continue

            for idx, config in enumerate(steps):
                if not isinstance(config, dict):
                    # should not happen
                    continue

                attributes = deepcopy(config)
                attributes[CustomAttributes.RESOURCE_TYPE] = ResourceType.STEPS

                block_name = f"{ResourceType.JOBS}.{name}.{ResourceType.STEPS}.{idx + 1}"

                block = Block(
                    name=block_name,
                    config=config,
                    path=file_path,
                    block_type=BlockType.RESOURCE,
                    attributes=attributes,
                    id=block_name,
                    source=self.source,
                )
                self.vertices.append(block)
                self.job_steps_map[(file_path, f"{ResourceType.JOBS}.{name}")].append((file_path, block_name))

    def _create_permissions_vertices(self, file_path: str, permissions: Any) -> None:
        """Creates root-level permissions vertices"""

        if permissions is None:
            # if 'permissions' is not set in a file, then it is automatically 'write-all'
            permissions = {
                "permissions": "write-all",
                START_LINE: 0,
                END_LINE: 0,
            }

        if not permissions or not isinstance(permissions, (str, dict)):
            return

        if isinstance(permissions, str):
            # to get the correct line numbers we would need to check the raw definition
            config = {
                "permissions": permissions,
                START_LINE: 0,
                END_LINE: 0,
            }
        else:
            config = permissions

        attributes = deepcopy(config)
        attributes[CustomAttributes.RESOURCE_TYPE] = ResourceType.PERMISSIONS

        block_name = ResourceType.PERMISSIONS

        block = Block(
            name=block_name,
            config=config,
            path=file_path,
            block_type=BlockType.RESOURCE,
            attributes=attributes,
            id=block_name,
            source=self.source,
        )
        self.vertices.append(block)

    def _create_on_vertices(self, file_path: str, on: Any) -> None:
        if not on:
            return

        if isinstance(on, (str, list)):
            # to get the correct line numbers we would need to check the raw definition
            config = {
                "on": on,
                START_LINE: 0,
                END_LINE: 0,
            }
        elif isinstance(on, dict):
            config = on
        else:
            return

        attributes = deepcopy(config)
        attributes[CustomAttributes.RESOURCE_TYPE] = ResourceType.ON

        block_name = ResourceType.ON

        block = Block(
            name=block_name,
            config=config,
            path=file_path,
            block_type=BlockType.RESOURCE,
            attributes=attributes,
            id=block_name,
            source=self.source,
        )
        self.vertices.append(block)

    def _create_edges(self) -> None:
        self._create_jobs_to_steps_edges()

    def _create_jobs_to_steps_edges(self) -> None:
        """Creates edges from jobs to steps"""

        if not self.job_steps_map:
            return

        for path_and_name, path_and_steps in self.job_steps_map.items():
            origin_vertex_index = self.vertices_by_path_and_name[path_and_name]

            for path_and_step_name in path_and_steps:
                dest_vertex_index = self.vertices_by_path_and_name[path_and_step_name]

                self._create_edge(
                    origin_vertex_index=origin_vertex_index,
                    dest_vertex_index=dest_vertex_index,
                )

    @staticmethod
    def get_files_definitions(root_folder: str | Path) -> dict[str | Path, dict[str, Any] | list[dict[str, Any]]]:
        definitions: "dict[str | Path, dict[str, Any] | list[dict[str, Any]]]" = {}
        file_paths = get_scannable_file_paths(root_folder=root_folder)

        for file_path in file_paths:
            result = parse_file(f=file_path)
            if result is not None:
                definitions[file_path] = result[0]

        return definitions



from checkov.common.models.enums import CheckCategories
from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceValueCheck


class FunctionAppHttpVersionLatest(BaseResourceValueCheck):
    def __init__(self):
        name = "Ensure that 'HTTP Version' is the latest, if used to run the Function app"
        id = "CKV_AZURE_67"
        supported_resources = ['azurerm_function_app', 'azurerm_function_app_slot']
        categories = [CheckCategories.GENERAL_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self):
        return 'site_config/[0]/http2_enabled'


check = FunctionAppHttpVersionLatest()



from __future__ import annotations

from checkov.common.models.enums import CheckResult
from checkov.github.base_github_negative_branch_security import NegativeBranchSecurity


class GithubRequire2Approvals(NegativeBranchSecurity):
    def __init__(self) -> None:
        name = "Ensure any change to code receives approval of two strongly authenticated users"
        id = "CKV_GITHUB_19"
        super().__init__(name=name, id=id, missing_attribute_result=CheckResult.FAILED)

    def get_evaluated_keys(self) -> list[str]:
        return ["required_pull_request_reviews/required_approving_review_count"]

    def get_forbidden_values(self) -> list[int | None]:
        return [None, 0, 1]


check = GithubRequire2Approvals()
